# shows some use cases with the Kalman filter class

import numpy as np
import matplotlib.pyplot as plt
from gsoup.track import KalmanFilter
from scipy import signal

# discretization step of simulation
dt = 0.1
# initial values for the simulation
initialPosition = 10
initialVelocity = -5
acceleration = 0.5
# measurement noise standard deviation
noiseStd = 0.01
# number of discretization time steps
numberTimeSteps = 1000
# for a more elaborate use case: number of lookahead future prediction steps
k = 10
# for a more elaborate use case: missing measurements stride
mask_interval = 10
gt_motion = "sawtooth"  # sine, constant_acc, sawtooth
# assume that for short durations, the motion is constant acceleration
A = np.matrix(
    [
        [1, dt, 0.5 * (dt**2)],
        [0, 1, dt],
        [0, 0, 1],
    ]
)  # state-change/transition matrix
B = np.matrix([[0], [0], [0]])  # control matrix
C = np.matrix([[1, 0, 0]])  # measurement/observation matrix
Q = 2500 * np.matrix(np.eye(3))  # process noise covariance matrix
R = (noiseStd**2) * np.matrix([[1]])  # measurement noise covariance matrix

x0 = np.matrix([[0], [0], [0]])  # guess of the initial estimate
P0 = 1 * np.matrix(np.eye(3))  # guess of the initial estimate error covariance matrix

############################################################################
############################################################################
############################################################################

timeVector = np.linspace(0, (numberTimeSteps - 1) * dt, numberTimeSteps)
mask = np.zeros(np.size(timeVector))
mask[::mask_interval] = 1
steps = np.arange(np.size(timeVector))
position = np.zeros(np.size(timeVector))
velocity = np.zeros(np.size(timeVector))
acceleration = np.zeros(np.size(timeVector))

# simulate the system behavior
for i in np.arange(np.size(timeVector)):
    if gt_motion == "sawtooth":
        amp = 5
        # lets use a saw tooth wave
        position[i] = amp * signal.sawtooth(i / 20, 0.5)
        # its derivative
        velocity[i] = 2 * np.pi * 5 / 20 * signal.square(i / 20, 0.5)
        # formally should take value 1, -1 in peaks...lets just use 0
        acceleration[i] = 0
    elif gt_motion == "sine":
        # sine wave
        position[i] = 5 * np.sin(0.1 * timeVector[i])
        velocity[i] = 0.1 * 5 * np.cos(0.1 * timeVector[i])
        acceleration[i] = 0.1 * 0.1 * 5 * -np.sin(0.1 * timeVector[i])
    elif gt_motion == "constant_acc":
        # constant acc
        position[i] = (
            initialPosition
            + initialVelocity * timeVector[i]
            + (acceleration * timeVector[i] ** 2) / 2
        )
        velocity[i] = initialVelocity + acceleration * timeVector[i]
        acceleration[i] = acceleration
if gt_motion == "sawtooth":
    acceleration = np.concatenate((np.diff(velocity), np.zeros(1)))
# add the measurement noise
positionNoisy = position + np.random.normal(0, noiseStd, size=np.size(timeVector))
# verify the position vector by plotting the results
# plotStep = numberTimeSteps
# plt.figure(figsize=(10, 10))
# plt.plot(steps[0:plotStep], position[0:plotStep], linewidth=4, label="Ideal position")
# plt.plot(steps[0:plotStep], positionNoisy[0:plotStep], "r", label="Observed position")
# plt.xlabel("time")
# plt.ylabel("position")
# plt.legend()
# plt.savefig("data.png", dpi=300)
# plt.show()


# create a Kalman filter object
KalmanFilterObject = KalmanFilter(x0, P0, A, B, C, Q, R)
inputValue = np.matrix([[0]])
xk_plus, pk_plus = x0, P0
masked_xk_plus, masked_pk_plus = x0, P0
predictions = []
estimates_k = []
estimates_masked = []
estimates_masked.append(x0)
estimates = []
estimates.append(x0)

for i in range(k):
    estimates_k.append(np.matrix([[0], [0], [0]]))
# simulate online prediction
for j in np.arange(np.size(timeVector)):
    print("Time step: {}".format(j))
    xk_minus_k = KalmanFilterObject.predict_multi(inputValue, k, xk_plus)
    estimates_k.append(xk_minus_k)
    xk_minus, pk_minus = KalmanFilterObject.predict(inputValue, xk_plus, pk_plus)
    predictions.append(xk_minus)
    masked_xk_minus, masked_pk_minus = KalmanFilterObject.predict(
        inputValue, masked_xk_plus, masked_pk_plus
    )
    if mask[j] == 0:
        masked_xk_plus = masked_xk_minus
        masked_pk_plus = masked_pk_minus
    else:
        masked_xk_plus, masked_pk_plus = KalmanFilterObject.update(
            positionNoisy[j], masked_xk_minus, masked_pk_minus
        )
    xk_plus, pk_plus = KalmanFilterObject.update(positionNoisy[j], xk_minus, pk_minus)
    estimates.append(xk_plus)
    estimates_masked.append(masked_xk_plus)


# extract the state estimates in order to plot the results
estimate1 = []
estimate2 = []
estimate3 = []
# k steps ahead
estimate1_k = []
estimate2_k = []
estimate3_k = []
# masked
estimate1_masked = []
estimate2_masked = []
estimate3_masked = []


for j in np.arange(np.size(timeVector)):
    # estimates n+1
    estimate1.append(estimates[j][0, 0])
    estimate2.append(estimates[j][1, 0])
    estimate3.append(estimates[j][2, 0])
    # estimates n+k
    estimate1_k.append(estimates_k[j][0, 0])
    estimate2_k.append(estimates_k[j][1, 0])
    estimate3_k.append(estimates_k[j][2, 0])
    # estimates n+k masked
    estimate1_masked.append(estimates_masked[j][0, 0])
    estimate2_masked.append(estimates_masked[j][1, 0])
    estimate3_masked.append(estimates_masked[j][2, 0])

# create vectors corresponding to the true values in order to plot the results
estimate1true = position
estimate2true = velocity
# estimate3true = acceleration * np.ones(np.size(timeVector))
estimate3true = acceleration


# plot the results

fig, ax = plt.subplots(3, 1, figsize=(10, 15))

ax[0].scatter(
    steps,
    estimate1_k,
    color="orange",
    s=3,
    # linestyle="-",
    label="k={} lookahead steps".format(k),
)
ax[0].scatter(
    steps,
    estimate1,
    color="blue",
    s=3,
    # linestyle="-",
    label="k=1 lookahead steps",
)
ax[0].scatter(
    steps,
    estimate1_masked,
    color="green",
    s=3,
    # linestyle="-",
    label="{}/{} missing measurements".format(mask_interval - 1, mask_interval),
)
ax[0].scatter(
    steps,
    estimate1true,
    color="red",
    s=3,
    # linestyle="-",
    label="ground truth",
)

ax[0].set_xlabel("Discrete-time steps k", fontsize=14)
ax[0].set_ylabel("Position", fontsize=14)
ax[0].tick_params(axis="both", labelsize=12)
ax[0].grid()
ax[0].legend(fontsize=14)
ax[1].plot(
    steps,
    estimate2_k,
    color="orange",
    linestyle="-",
    label="k={} lookahead steps".format(k),
)
ax[1].plot(
    steps,
    estimate2,
    color="blue",
    linestyle="-",
    label="k=1 lookahead steps",
)
ax[1].plot(
    steps,
    estimate2_masked,
    color="green",
    linestyle="-",
    label="{}/{} missing measurements".format(mask_interval - 1, mask_interval),
)
ax[1].plot(
    steps,
    estimate2true,
    color="red",
    linestyle="-",
    label="ground truth",
)

ax[1].set_xlabel("Discrete-time steps k", fontsize=14)
ax[1].set_ylabel("Velocity", fontsize=14)
ax[1].tick_params(axis="both", labelsize=12)
ax[1].grid()
ax[1].legend(fontsize=14)
ax[2].plot(
    steps,
    estimate3_k,
    color="orange",
    linestyle="-",
    label="k={} lookahead steps".format(k),
)
ax[2].plot(
    steps,
    estimate3,
    color="blue",
    linestyle="-",
    label="k=1 lookahead steps",
)
ax[2].plot(
    steps,
    estimate3_masked,
    color="green",
    linestyle="-",
    label="{}/{} missing measurements".format(mask_interval - 1, mask_interval),
)
ax[2].plot(
    steps,
    estimate3true,
    color="red",
    linestyle="-",
    label="ground truth",
)

ax[2].set_xlabel("Discrete-time steps k", fontsize=14)
ax[2].set_ylabel("Acceleration", fontsize=14)
ax[2].tick_params(axis="both", labelsize=12)
ax[2].grid()
ax[2].legend(fontsize=14)
fig.savefig("plots.png", dpi=600)



__version__ = "0.2.6"

from .core import (
    is_np,
    to_hom,
    homogenize,
    compose_rt,
    to_44,
    to_34,
    to_np,
    to_numpy,
    to_torch,
    to_8b,
    to_float,
    color_to_gray,
    broadcast_batch,
    repeat,
    map_range,
    map_to_01,
    swap_columns,
)

from .transforms import (
    rotx,
    roty,
    rotz,
    translate,
    scale,
    rotate,
    invert_rigid,
    look_at_torch,
    look_at_np,
    create_random_cameras_on_unit_sphere,
    opengl_c2w_to_opencv_c2w,
    opencv_c2w_to_opengl_c2w,
    opencv_intrinsics_from_opengl_project,
    opengl_project_from_opencv_intrinsics,
    perspective_projection,
    pixels_in_world_space,
    find_rigid_transform,
    find_affine_transformation,
    decompose_affine,
    random_vectors_on_sphere,
    random_affine,
    random_perspective,
    random_qvec,
    vec2skew,
    batch_vec2skew,
    rotvec2mat,
    batch_rotvec2mat,
    qvec2mat,
    mat2qvec,
)
from .geometry_basic import (
    project_point_to_line,
    project_point_to_segment,
    point_line_distance,
    is_inside_triangle,
    triangulate_quad_mesh,
    ray_ray_intersection,
    duplicate_faces,
    remove_duplicate_faces,
    get_aspect_ratio,
    normalize_vertices,
    clean_infinite_vertices,
    faces2edges_naive,
    calc_edges,
    calc_face_normals,
    calc_vertex_normals,
    calc_edge_length,
    get_center_of_attention,
    scale_poses,
    find_princple_componenets,
    qslerp,
)

from .geometry_advanced import (
    distribute_field,
    distribute_scalar_field,
    distribute_vector_field,
    qem,
)

from .gsoup_io import (
    write_to_json,
    save_image,
    save_images,
    save_animation,
    save_mesh,
    save_meshes,
    load_image,
    load_images,
    load_mesh,
    save_pointcloud,
    save_pointclouds,
    load_pointcloud,
    read_exr,
    write_exr,
)

from .image import (
    add_alpha,
    alpha_compose,
    draw_text_on_image,
    draw_gizmo_on_image,
    merge_figures_with_line,
    generate_random_block_mask,
    generate_checkerboard,
    generate_voronoi_diagram,
    generate_gray_gradient,
    generate_dot_pattern,
    generate_stripe_pattern,
    generate_concentric_circles,
    generate_lollipop_pattern,
    generate_gaussian_image,
    image_grid,
    resize,
    resize_images_naive,
    adjust_contrast_brightness,
    change_brightness,
    linear_to_srgb,
    srgb_to_linear,
    pad_to_res,
    pad_to_square,
    crop_to_square,
    mask_regions,
    crop_center,
    compute_color_distance,
    tonemap,
)

# from .video import (
#     get_ffmpeg_version,
#     get_video_info,
#     get_frame_timestamps,
#     load_video,
#     save_video,
#     reverse_video,
#     compress_video,
#     video_to_images,
#     images_to_video,
#     slice_from_video,
#     trim_video,
#     VideoReader,
#     FPS,
# )

from .video import (
    probe_video,
    load_video,
    save_video,
    get_single_frame,
    get_frame_timestamps,
    reverse_video,
    compress_video,
    slice_video,
    trim_video,
    VideoReader,
    FPS,
)

from .procam import (
    warp_image,
    calibrate_procam,
    compute_backward_map,
    naive_color_compensate,
    reconstruct_pointcloud,
    GrayCode,
)

from .sphere_trace import (
    generate_rays,
    render_sdf,
)

from .rasterize import (
    render_mesh,
    project_points,
    should_cull_tri,
)

from . import structures



import numpy as np
from .core import to_hom, to_44
from .transforms import invert_rigid
from .geometry_basic import triangulate_quad_mesh


def barycentric(p, a, b, c):
    """
    get 2D barycentric coordinates of 2D point p for triangle defined by endpoints a,b,c
    :param p: (2,) np.float32 coordinates of point in same coordinate system as a,b,c
    :param a,b,c: (3,) np.float32 end points of triangle
    :return: barycentric coordinates u,v,w of p
    """
    v0, v1, v2 = b - a, c - a, p - a
    d00, d01, d11 = np.dot(v0, v0), np.dot(v0, v1), np.dot(v1, v1)
    d20, d21 = np.dot(v2, v0), np.dot(v2, v1)
    denom = d00 * d11 - d01 * d01
    v = (d11 * d20 - d01 * d21) / denom
    w = (d00 * d21 - d01 * d20) / denom
    u = 1.0 - v - w
    return u, v, w


def is_inside_triangle(p, a, b, c):
    """
    tests if p is inside triangle defined by endpoints a,b,c
    :return: True if p is inside, and the barycentric coordinates of p
    """
    u, v, w = barycentric(p, a, b, c)
    return u >= 0 and v >= 0 and w >= 0, u, v, w


def project_points(points, K, w2c):
    """
    projects 3D points to camera screen space
    :param points: (n, 3) np.float32 points to be projected
    :param K: (3, 3) np.float32 intrinsics of camera
    :param w2c: (3, 4) np.float32 extrinsics of camera
    """
    points_h = to_hom(points)
    projected = K @ w2c @ points_h.T  # Apply camera projection
    projected = projected.T
    projected /= projected[:, 2:]  # Normalize by depth
    return projected  # second dim is (x, y, depth)


def should_cull_tri(v0, v1, v2, camera_pos):
    """
    a slightly naive cull procedure (should use camera view direction)
    """
    normal = np.cross(v1 - v0, v2 - v0)
    view_dir = v0 - camera_pos
    return np.dot(normal, view_dir) >= 0  # Cull back-facing


def draw_line(image, p0, p1, color):
    """
    Draws a line between p0 and p1 on the image using Bresenham's algorithm.

    :param image: (height, width, 3) np.uint8 image.
    :param p0: Starting point (x, y).
    :param p1: Ending point (x, y).
    :param color: (3,) np.uint8 color.
    """
    x0, y0 = np.round(p0).astype(np.int32)
    x1, y1 = np.round(p1).astype(np.int32)
    dx = abs(x1 - x0)
    dy = abs(y1 - y0)
    sx = 1 if x0 < x1 else -1
    sy = 1 if y0 < y1 else -1
    err = dx - dy

    while True:
        if 0 <= x0 < image.shape[1] and 0 <= y0 < image.shape[0]:
            image[y0, x0] = color
        if x0 == x1 and y0 == y1:
            break
        e2 = 2 * err
        if e2 > -dy:
            err -= dy
            x0 += sx
        if e2 < dx:
            err += dx
            y0 += sy


def draw_triangle(image, depth_buffer, a, b, c, color):
    """
    draws a triangle a,b,c into image
    """
    height, width = image.shape[:2]
    min_x, max_x = max(0, min(a[0], b[0], c[0])), min(width - 1, max(a[0], b[0], c[0]))
    min_y, max_y = max(0, min(a[1], b[1], c[1])), min(height - 1, max(a[1], b[1], c[1]))
    for y in range(int(min_y), int(max_y + 1)):
        for x in range(int(min_x), int(max_x + 1)):
            if 0 <= x < width and 0 <= y < height:
                inside, u, v, w = is_inside_triangle(
                    np.array([x, y]), a[:2], b[:2], c[:2]
                )
                if inside:
                    depth = u * a[2] + v * b[2] + w * c[2]
                    if depth < depth_buffer[y, x]:
                        depth_buffer[y, x] = depth
                        image[y, x] = color


def render_wireframe(image, V, F, K, w2c, color, wireframe_occlude):
    """
    Renders the wireframe of a mesh. Supports both triangle and quad faces.

    :param image: (height, width, 3) np.uint8 image to draw on.
    :param V: (n, 3) np.float32 vertices in world coordinates.
    :param F: list or array of face indices (each face can be a triangle [3] or a quad [4]).
    :param K: (3, 3) np.float32 camera intrinsics.
    :param w2c: (3, 4) np.float32 camera extrinsics (world -> cam).
    :param color: (m, 3) np.uint8 line color per face.
    :param wireframe_occlude: if true, will not render wireframe on backfaces
    """
    # Project the vertices to screen space.
    projected_vertices = project_points(V, K, w2c)[:, :2]
    for f in F:
        if wireframe_occlude:
            v0, v1, v2 = V[f[0]], V[f[1]], V[f[2]]
            camera_pos = invert_rigid(to_44(w2c)[None, :])[0, :3, -1]  # get cam pose
            if should_cull_tri(v0, v1, v2, camera_pos):
                continue
        # Draw an edge from each vertex to the next, wrapping around.
        n = len(f)
        for i in range(n):
            p0 = projected_vertices[f[i]]
            p1 = projected_vertices[f[(i + 1) % n]]
            draw_line(image, p0, p1, color[i])


def render_mesh(
    V,
    F,
    K,
    w2c,
    color,
    wireframe=False,
    wireframe_occlude=False,
    image=None,
    depth_buffer=None,
    wh=(512, 512),
):
    """
    renders a mesh onto an image
    :param V: a (n, 3) np.float32 of vertices of a mesh in world coordinates
    :param F: a (m, 3) or (m, 4) np.int32 indices into V, defining the faces of the mesh (assumes CCW order)
    :param K: a (3, 3) np.float32 intrinsics matrix (opencv convention)
    :param w2c: a (3, 4) np.float32 extrinsics matrix (opencv convention, world -> cam)
    :param color: (m, 3) np.uint8 color per face, or (3,) for single color.
    :param wireframe: if True, will render the wireframe version of the mesh
    :param wireframe_occlude: if True, will not render wireframe on backfaces
    :param image: if not None, a (height, width 3) np.uint8 to be rendered into
    :param depth_buffer: if not None, a (height, width) np.float32 to store depth and perform z-testing
    :param wh: (2-tuple) width height to use if image/depthbuffer aren't provided
    :return (w, h, 3) np.uint8 rendered image
    """
    projected_vertices = project_points(V, K, w2c)  # project vertices to screen space
    if image is None:
        image = np.zeros((wh[1], wh[0], 3), dtype=np.uint8)
    else:
        image = image.copy()
    if depth_buffer is None and not wireframe:
        depth_buffer = np.full((wh[1], wh[0]), np.inf, dtype=np.float32)
    if image.shape[0:2] != image.shape[0:2]:
        raise ValueError("image and depth buffer must have same spatial dimensions")
    if wireframe:
        if color.ndim == 1:  # handle single color
            color = np.tile(color[None, :], (len(F), 1))
        render_wireframe(
            image,
            V,
            F,
            K,
            w2c,
            color,
            wireframe_occlude=wireframe_occlude,
        )
    else:
        if F.shape[-1] == 4:  # triangulate for rendering
            F, color = triangulate_quad_mesh(F, color)
        if color.ndim == 1:  # handle single color
            color = np.tile(color[None, :], len(F))
        for i, f in enumerate(F):
            v0, v1, v2 = V[f[0]], V[f[1]], V[f[2]]
            camera_pos = invert_rigid(to_44(w2c)[None, :])[0, :3, -1]  # get cam pose
            if not should_cull_tri(v0, v1, v2, camera_pos):
                draw_triangle(
                    image,
                    depth_buffer,
                    projected_vertices[f[0]],
                    projected_vertices[f[1]],
                    projected_vertices[f[2]],
                    color[i],
                )
    return image



import pytest
import os
import numpy as np
import gsoup
from pathlib import Path


@pytest.fixture
def dummy_video():
    """
    Create a dummy video file with 10 frames (64x64) where each frame is filled
    with a constant value that increases with the frame index.
    """
    n_frames = 10
    height, width = 64, 64
    frames = []
    for i in range(n_frames):
        # Create a frame with a constant value; values increase with frame index.
        frame = np.full((height, width, 3), fill_value=i * 25, dtype=np.uint8)
        frames.append(frame)
    frames_np = np.array(frames)
    video_path = Path("resource", "dummy.mp4")
    # Write video at 10 fps with a given bitrate.
    gsoup.save_video(frames_np, video_path, fps=10, bitrate="500k")
    return video_path, frames_np


@pytest.fixture
def dummy_video_long():
    """
    Create a dummy video file with 10 frames (64x64) where each frame is filled
    with a constant value that increases with the frame index.
    """
    n_frames = 1000
    height, width = 64, 64
    frames = []
    for i in range(n_frames):
        # Create a frame with a constant value; values increase with frame index.
        frame = np.random.randint(0, high=256, size=(height, width, 3))
        frames.append(frame)
    frames_np = np.array(frames)
    video_path = Path("resource", "dummy_long.mp4")
    # Write video at 10 fps with a given bitrate.
    gsoup.save_video(frames_np, video_path, fps=10, bitrate="500k")
    return video_path, frames_np


def test_probe_video(dummy_video):
    video_path, _ = dummy_video
    info = gsoup.probe_video(video_path)
    # Check that width and height are detected correctly.
    assert info["width"] == 64
    assert info["height"] == 64
    # The number of frames might be approximate due to encoding;
    # ensure we have at least most of them.
    assert int(info["nb_frames"]) >= 8


def test_load_video(dummy_video):
    video_path, frames_np = dummy_video
    # Test reading all frames at once.
    frames = gsoup.load_video(video_path)
    assert frames.ndim == 4
    assert len(frames) == frames_np.shape[0]
    assert frames.shape[1] == 64
    assert frames.shape[2] == 64
    # Test streaming version.
    # frames_stream = list(gsoup.load_video(video_path, stream=True))


def test_get_frame(dummy_video):
    video_path, frames_np = dummy_video
    # Extract the first frame.
    frame = gsoup.get_single_frame(video_path, 0)
    assert frame.shape == (64, 64, 3)
    # Due to potential compression artifacts, allow a small tolerance.
    diff = np.abs(frame.astype(np.int32) - frames_np[0].astype(np.int32))
    assert np.mean(diff) < 20


def test_reverse_video(dummy_video):
    video_path, frames_np = dummy_video
    reversed_video_path = Path("resource", "reversed.mp4")
    gsoup.reverse_video(video_path, reversed_video_path, bitrate="500k")
    # Read frames from the reversed video.
    rev_frames = gsoup.load_video(reversed_video_path)
    # Compare with the original frames in reverse order.
    original_reversed = frames_np[::-1]
    min_frames = min(len(rev_frames), len(original_reversed))
    diff = np.abs(
        rev_frames[:min_frames].astype(np.int32)
        - original_reversed[:min_frames].astype(np.int32)
    )
    assert np.mean(diff) < 25


def test_compress_video(dummy_video):
    video_path, _ = dummy_video
    compressed_video_path = Path("resource", "compressed.mp4")
    gsoup.compress_video(video_path, compressed_video_path, bitrate="100k")
    assert compressed_video_path.exists()
    # Optionally, compare file sizes (compressed should be smaller).
    orig_size = os.path.getsize(video_path)
    comp_size = os.path.getsize(compressed_video_path)
    assert comp_size < orig_size


def test_slice_video(dummy_video):
    video_path, _ = dummy_video
    # Slice frames from 2 to 8 with a stride of 2 (expecting frames 2, 4, 6, 8).
    sliced_frames = gsoup.slice_video(
        video_path,
        start_frame=2,
        stop_frame=8,
        stride=2,
        fps=10,
        bitrate="500k",
    )
    assert sliced_frames.shape[0] == 4
    sliced_video_path = Path("resource", "sliced.mp4")
    gsoup.slice_video(
        video_path,
        start_frame=2,
        stop_frame=8,
        stride=2,
        fps=10,
        bitrate="500k",
        to_numpy=False,
        output_path=sliced_video_path,
    )
    sliced_frames = gsoup.load_video(sliced_video_path)
    assert sliced_frames.shape[0] == 4


def test_trim_video(dummy_video):
    video_path, _ = dummy_video
    # Trim video from frame 3 to frame 7 (should contain ~4 frames).
    trimmed_video_path = Path("resource", "trimmed.mp4")
    gsoup.trim_video(
        video_path, trimmed_video_path, start_frame=3, end_frame=7, bitrate="500k"
    )
    trimmed_frames = gsoup.load_video(trimmed_video_path)
    # Allow a one-frame tolerance.
    assert abs(trimmed_frames.shape[0] - 4) <= 1


def test_get_frame_timestamps(dummy_video):
    video_path, _ = dummy_video
    timestamps = gsoup.get_frame_timestamps(video_path)
    assert np.all(np.isclose(timestamps, np.linspace(0.0, 1.0, 10, endpoint=False)))


def test_video_reader(dummy_video_long):
    video_path, frames_np = dummy_video_long
    reader = gsoup.VideoReader(video_path)
    # fps = gsoup.FPS()
    frames = []
    for frame in reader:
        frames.append(frame)
        # print(fps())
    assert len(frames) == frames_np.shape[0]



import torch
import torch.nn as nn


def sphere_tracing(
    signed_distance_function,
    ray_positions,
    ray_directions,
    num_iterations,
    convergence_threshold,
    foreground_masks=None,
    bounding_radius=None,
    count_access=False,
):
    """
    Sphere trace a SDF in pytorch.
    note: can be made differentiable using https://arxiv.org/abs/1912.07372
    :param signed_distance_function: a function that takes a tensor of shape (N, 3) and returns a tensor of shape (N, 1)
    :param ray_positions: a tensor of shape (N, 3)
    :param ray_directions: a tensor of shape (N, 3)
    :param num_iterations: the number of iterations to run
    :param convergence_threshold: the threshold from the surface for a ray to be considered converged
    :param foreground_masks: a tensor of shape (N, 1) that indicates which rays are foreground
    :param bounding_radius: a float that indicates the radius of the scene bounding sphere
    :param count_access: a boolean that indicates whether to count the number of times the SDF is accessed
    """
    counter = 0
    if foreground_masks is None:
        foreground_masks = torch.all(
            torch.isfinite(ray_positions), dim=-1, keepdim=True
        )
    if bounding_radius:
        a = torch.sum(ray_directions * ray_directions, dim=-1, keepdim=True)
        b = 2 * torch.sum(ray_directions * ray_positions, dim=-1, keepdim=True)
        c = (
            torch.sum(ray_positions * ray_positions, dim=-1, keepdim=True)
            - bounding_radius**2
        )
        d = b**2 - 4 * a * c
        t = (-b - torch.sqrt(d)) / (2 * a)
        bounded = d >= 0
        ray_positions = torch.where(
            bounded, ray_positions + ray_directions * t, ray_positions
        )
        foreground_masks = foreground_masks & bounded
    foreground_masks = foreground_masks[:, 0]
    with torch.no_grad():
        converged = torch.zeros(
            (ray_positions.shape[:-1]), device=ray_positions.device, dtype=torch.bool
        )
        for _ in range(num_iterations):
            mask = foreground_masks & ~converged
            cur_ray_positions = ray_positions.view(-1, 3)[mask]
            cur_ray_directions = ray_directions.view(-1, 3)[mask]
            signed_distances = signed_distance_function(cur_ray_positions).view(-1, 1)
            if count_access:
                counter += cur_ray_positions.shape[0]
            cur_ray_positions = (
                cur_ray_positions + cur_ray_directions * signed_distances
            )
            ray_positions[mask] = cur_ray_positions
            if bounding_radius:
                bounded = torch.norm(ray_positions, dim=-1) < bounding_radius
                foreground_masks = foreground_masks & bounded
            converged[mask] |= torch.abs(signed_distances[:, 0]) < convergence_threshold
            if torch.all(~foreground_masks | converged):
                break
    return ray_positions, converged[:, None], counter


def compute_shadows(
    signed_distance_function,
    surface_positions,
    surface_normals,
    light_directions,
    num_iterations,
    convergence_threshold,
    foreground_masks=None,
    bounding_radius=None,
):
    surface_positions, converged, _ = sphere_tracing(
        signed_distance_function=signed_distance_function,
        ray_positions=surface_positions + surface_normals * 1e-3,
        ray_directions=light_directions,
        num_iterations=num_iterations,
        convergence_threshold=convergence_threshold,
        foreground_masks=foreground_masks,
        bounding_radius=bounding_radius,
    )
    return foreground_masks & converged


def compute_normal(
    signed_distance_function,
    surface_positions,
    finite_difference_epsilon,
    use_gradient=False,
):
    if use_gradient:
        surface_positions.requires_grad = True
        raw = signed_distance_function(surface_positions)
        d_output = torch.ones_like(raw, requires_grad=False, device=raw.device)
        surface_normals = torch.autograd.grad(
            outputs=raw,
            inputs=surface_positions,
            grad_outputs=d_output,
            create_graph=False,
            retain_graph=False,
        )[0]
    else:
        finite_difference_epsilon_x = surface_positions.new_tensor(
            [finite_difference_epsilon, 0.0, 0.0]
        )
        finite_difference_epsilon_y = surface_positions.new_tensor(
            [0.0, finite_difference_epsilon, 0.0]
        )
        finite_difference_epsilon_z = surface_positions.new_tensor(
            [0.0, 0.0, finite_difference_epsilon]
        )
        surface_normals_x = signed_distance_function(
            surface_positions + finite_difference_epsilon_x
        ) - signed_distance_function(surface_positions - finite_difference_epsilon_x)
        surface_normals_y = signed_distance_function(
            surface_positions + finite_difference_epsilon_y
        ) - signed_distance_function(surface_positions - finite_difference_epsilon_y)
        surface_normals_z = signed_distance_function(
            surface_positions + finite_difference_epsilon_z
        ) - signed_distance_function(surface_positions - finite_difference_epsilon_z)
        surface_normals = torch.cat(
            (surface_normals_x, surface_normals_y, surface_normals_z), dim=-1
        )
    surface_normals = nn.functional.normalize(surface_normals, dim=-1)
    return surface_normals


def generate_rays(w2v, v2c, resx=512, resy=512, device="cuda:0"):
    """
    generates batch of rays for a batch of cameras
    :param w2v: a batch of world to view matrices (nx4x4)
    :param v2c: view to clip matrix (4x4)
    :param resx: resolution x
    :param resy: resolution y
    :param device: device
    """
    y_clip = (torch.arange(resy, dtype=torch.float32, device=device) / resy) * 2 - 1
    x_clip = (torch.arange(resx, dtype=torch.float32, device=device) / resx) * 2 - 1
    xy_clip = torch.stack(torch.meshgrid(x_clip, y_clip, indexing="ij"), dim=-1)
    xy_clip_near = torch.cat(
        (
            xy_clip,
            -1 * torch.ones_like(xy_clip[:, :, :1]),
            torch.ones_like(xy_clip[:, :, :1]),
        ),
        dim=-1,
    )
    xy_clip_far = torch.cat(
        (
            xy_clip,
            torch.ones_like(xy_clip[:, :, :1]),
            torch.ones_like(xy_clip[:, :, :1]),
        ),
        dim=-1,
    )
    xy_view_near = (torch.inverse(v2c) @ xy_clip_near.view(-1, 4).T).T.view(
        resy, resx, 4
    )
    xy_view_near = xy_view_near / xy_view_near[:, :, 3:]
    rays_d = []
    rays_o = []
    for i in range(len(w2v)):
        if v2c.ndim == 3:
            v2c_transform = v2c[i]
        else:
            v2c_transform = v2c
        w2v_transform = w2v[i]
        xy_view_near = (
            torch.inverse(v2c_transform) @ xy_clip_near.view(-1, 4).T
        ).T.view(resy, resx, 4)
        xy_view_near = xy_view_near / xy_view_near[:, :, 3:]
        xy_world_near = (
            torch.inverse(w2v_transform) @ xy_view_near.view(-1, 4).T
        ).T.view(resy, resx, 4)
        ray_o = xy_world_near[:, :, :3]
        camera_position = torch.inverse(w2v_transform)[:3, 3]
        ray_d = nn.functional.normalize(
            xy_world_near[:, :, :3] - camera_position, dim=-1
        )
        rays_d.append(ray_d)
        rays_o.append(ray_o)
    return torch.stack(rays_o), torch.stack(rays_d)


def render_sdf(
    p_sdf,
    ray_positions,
    ray_directions,
    num_iterations=2000,
    convergence_threshold=1e-4,
    use_gradient=False,
):
    surface_positions, converged, _ = sphere_tracing(
        signed_distance_function=p_sdf,
        ray_positions=ray_positions,
        ray_directions=ray_directions,
        num_iterations=num_iterations,
        convergence_threshold=convergence_threshold,
        bounding_radius=2.0,
    )
    surface_positions = torch.where(
        converged, surface_positions, torch.zeros_like(surface_positions)
    )
    surface_normals = compute_normal(
        signed_distance_function=p_sdf,
        surface_positions=surface_positions,
        finite_difference_epsilon=1e-4,
        use_gradient=use_gradient,
    )
    surface_normals = torch.where(
        converged, surface_normals, torch.zeros_like(surface_normals)
    )
    image = (surface_normals + 1.0) / 2.0
    image = torch.where(converged, image, torch.zeros_like(image))
    image = torch.concat((image, converged), dim=-1)
    return image



import time
import collections
import numpy as np
import pathlib
from pathlib import Path
import ffmpeg
import subprocess
import cv2

# import os
# FFMPEG_PATH = ""


def get_ffmpeg_version(verbose=False):
    """
    :return: ffmpeg version
    """
    # global FFMPEG_PATH
    FFMPEG_PATH = ""
    version = None
    # env_var = os.environ.get("FFMPEG_PATH")
    # if env_var:
    #     FFMPEG_PATH = env_var
    try:
        ffmpeg_output = subprocess.run(
            [str(Path(FFMPEG_PATH, "ffmpeg")), "-version"],
            capture_output=True,
            text=True,
        ).stdout
        parts = ffmpeg_output.split()
        version = parts[parts.index("version") + 1]
        if verbose:
            if int(version[0]) < 5:
                print(
                    "gsoup warning: detected ffmpeg version: {}. Video module May fail for versions lower than 5".format(
                        version
                    )
                )
                # print(
                #     "you can set the environment variable "
                #     "FFMPEG_PATH"
                #     " to the path of ffmpeg binary"
                # )
    except ValueError:
        if verbose:
            print(
                "gsoup warning: could not detect ffmpeg version. Video module May fail"
            )
            # print(
            #     "you can set the environment variable "
            #     "FFMPEG_PATH"
            #     " to the path of ffmpeg binary"
            # )
    except FileNotFoundError:
        if verbose:
            print("gsoup warning: ffmpeg not found. Video module May fail")
            # print(
            #     "you can set the environment variable "
            #     "FFMPEG_PATH"
            #     " to the path of ffmpeg binary"
            # )
    return version


get_ffmpeg_version(True)


class FPS:
    """
    calculates current fps and returns it, see https://stackoverflow.com/a/54539292
    example usage:
        fps = FPS()
        while True:
            # do something
            print(fps())
    """

    def __init__(self, avarageof=50):
        self.frametimestamps = collections.deque(maxlen=avarageof)

    def __call__(self):
        self.frametimestamps.append(time.time())
        nominator = len(self.frametimestamps)
        denominator = self.frametimestamps[-1] - self.frametimestamps[0]
        if (len(self.frametimestamps) > 1) and (denominator != 0):
            return nominator / denominator
        else:
            return 0.0


def get_video_info(video_path):
    """
    returns basic video info
    :param video_path: path to video
    :return: height, width, fps, frame_count
    """
    video_path = Path(video_path)
    if not video_path.exists():
        raise FileNotFoundError("Video file not found: {}".format(video_path))
    probe = ffmpeg.probe(str(video_path))
    video_stream = next(
        (stream for stream in probe["streams"] if stream["codec_type"] == "video"), None
    )
    width = int(video_stream["width"])
    height = int(video_stream["height"])
    fps = float(video_stream["r_frame_rate"].split("/")[0]) / float(
        video_stream["r_frame_rate"].split("/")[1]
    )
    frame_count = int(video_stream["nb_frames"])
    info = {
        "height": height,
        "width": width,
        "pixel_format": video_stream["pix_fmt"],
        "fps": fps,
        "frame_count": frame_count,
        "codec_name": video_stream["codec_name"],
        "bit_rate": video_stream["bit_rate"],
    }
    return info


def get_frame_timestamps(video_path):
    """
    returns numpy array of frame timestamps in seconds for a video
    :param video_path: path to video
    :return: numpy array of frame timestamps in seconds
    """
    video_path = Path(video_path)
    probe = ffmpeg.probe(str(video_path), show_frames="-show_frames")
    video_frames = [
        frame for frame in probe["frames"] if frame["media_type"] == "video"
    ]
    frame_times = np.array([float(frame["pts_time"]) for frame in video_frames])
    return frame_times


def load_video(video_path, verbose=False):
    """
    loads a video from disk into a numpy array (uint8, channels last, RGB)
    :param video_path: path to video
    :return: (n x h x w x 3) numpy array
    """
    info = get_video_info(video_path)
    out, _ = (
        ffmpeg.input(str(video_path))
        .output("pipe:", format="rawvideo", pix_fmt="rgb24")
        .run(capture_stdout=True, quiet=not verbose)
    )
    video = np.frombuffer(out, np.uint8).reshape([-1, info["height"], info["width"], 3])
    return video


def save_video(frames, output_path, fps, bit_rate="1M", lossy=True, verbose=False):
    """
    saves a video from a t x h x w x 3 numpy array
    :param frames: (t x h x w x 3) numpy array or directory path containing images of same format and resolution
    :param output_path: path to save video to
    :param fps: frames per second of output video
    :param lossy: if true, use lossy compression (if False, only .avi is supported)
    """
    output_path = Path(output_path)
    output_path.parent.mkdir(parents=True, exist_ok=True)
    if type(frames) == np.ndarray:
        if lossy:
            (
                ffmpeg.input(
                    "pipe:",
                    format="rawvideo",
                    pix_fmt="rgb24",
                    s="{}x{}".format(frames.shape[2], frames.shape[1]),
                    r=fps,
                )
                .output(str(output_path), pix_fmt="yuv420p", b=bit_rate)
                .overwrite_output()
                .run(input=frames.tobytes(), quiet=not verbose)
            )
        else:
            if output_path.suffix == ".avi":
                pix_fmt = "bgr24"
            else:
                # todo: figure out lossless pixel formats for other containers
                raise ValueError("Lossless video only supported for .avi container")
            (
                ffmpeg.input(
                    "pipe:",
                    format="rawvideo",
                    pix_fmt="rgb24",
                    s="{}x{}".format(frames.shape[2], frames.shape[1]),
                    r=fps,
                )
                .output(str(output_path), vcodec="rawvideo", pix_fmt=pix_fmt)
                .overwrite_output()
                .run(input=frames.tobytes(), quiet=not verbose)
            )
    elif isinstance(frames, pathlib.PurePath):
        if not frames.exists():
            raise FileNotFoundError("Image directory not found: {}".format(frames))
        if not frames.is_dir():
            raise ValueError(
                "frames must be numpy array or path to directory of images"
            )
        files = sorted(list(frames.glob("*")))
        if len(files) == 0:
            raise FileNotFoundError("No images found in directory: {}".format(frames))
        files = [x for x in files if x.is_file()]
        files = [x for x in files if x.suffix in [".png", ".jpg", ".jpeg", ".bmp"]]
        extensions = [file.suffix for file in files]
        if len(set(extensions)) > 1:
            raise ValueError("All images in directory must have same extension")
        with open("ffmpeg_input.txt", "wb") as outfile:
            for filename in files:
                mystr = "file '{}'\n".format(str(filename.resolve()).replace("\\", "/"))
                outfile.write(mystr.encode())
                mystr = "duration {:.05f}\n".format(1 / fps)
                outfile.write(mystr.encode())
        stdin_stream = None
        stdout_stream = subprocess.PIPE if not verbose else None
        stderr_stream = subprocess.PIPE if not verbose else None
        if lossy:
            args = [
                "ffmpeg",
                "-y",
                "-f",
                "concat",
                "-safe",
                "0",
                "-i",
                "ffmpeg_input.txt",
                "-b",
                bit_rate,
                "-r",
                str(fps),
                str(output_path),
            ]
        else:
            if output_path.suffix == ".avi":
                pix_fmt = "bgr24"
            else:
                # todo: figure out lossless pixel formats for other containers
                raise ValueError("Lossless video only supported for .avi container")
            args = [
                "ffmpeg",
                "-y",
                "-f",
                "concat",
                "-safe",
                "0",
                "-i",
                "ffmpeg_input.txt",
                "-r",
                str(fps),
                "-pix_fmt",
                pix_fmt,
                "-vcodec",
                "rawvideo",
                str(output_path),
            ]
        proc = subprocess.Popen(
            args, stdin=stdin_stream, stdout=stdout_stream, stderr=stderr_stream
        )
        out, err = proc.communicate(input)
        retcode = proc.poll()
        if retcode:
            raise ValueError("ffmpeg", out, err)
        Path("ffmpeg_input.txt").unlink()
    else:
        raise ValueError("frames must be numpy array or path to directory of images")


def reverse_video(video_path, output_path=None, verbose=False):
    """
    reverses a video and optionally saves it to disk
    :param video_path: path to video
    :param output_path: path to save video to
    :return: (n x h x w x 3) tensor of reversed video
    """
    info = get_video_info(video_path)
    out, _ = (
        ffmpeg.input(str(video_path))
        .filter("reverse")
        .output("pipe:", format="rawvideo", pix_fmt="rgb24")
        .run(capture_stdout=True, quiet=not verbose)
    )
    video = np.frombuffer(out, np.uint8).reshape([-1, info["height"], info["width"], 3])
    if output_path is not None:
        save_video(video, output_path)
    return video


def compress_video(src, dst, crf=23, verbose=False):
    """
    compresses a video using ffmpeg & libx264, saving it to disk
    :param src: path to video
    :param dst: path to save compressed video to
    :param crf: constant rate factor (lower is better quality, 0-51, 0 is lossless)
    :param verbose: if True, print ffmpeg output
    """
    if not (0 <= crf <= 51):
        raise ValueError("crf must be between 0 and 51")
    dst = Path(dst)
    dst.parent.mkdir(parents=True, exist_ok=True)
    (
        ffmpeg.input(str(src))
        .output(str(dst), vcodec="libx264", crf=crf, preset="slow")
        .overwrite_output()
        .run(quiet=not verbose)
    )


def images_to_video(
    src: Path, dst: Path, fps, bit_rate="1M", lossy=True, verbose=False
):
    """
    creates a video from a folder of images
    :param src: path to images
    :param dst: path to save video to
    :param fps: frames per second of output video
    :param bit_rate: bit rate of output video
    :param lossy: if true, use lossy compression (if False, only .avi is supported)
    :param verbose: if True, print ffmpeg output
    """
    save_video(src, dst, fps=fps, bit_rate=bit_rate, lossy=lossy, verbose=verbose)


def video_to_images(src, dst, verbose=False):
    """
    creates a folder of images from a video
    :param src: path to video
    :param dst: path to save images to (images will be named 0000.png, 0001.png, etc.)
    :param verbose: if True, print ffmpeg output
    """
    dst = Path(dst)
    dst.mkdir(parents=True, exist_ok=True)
    (
        ffmpeg.input(str(src))
        .output(str(dst / "%04d.png"), vcodec="png", format="image2")
        .overwrite_output()
        .run(quiet=not verbose)
    )


def slice_from_video(
    src, every_n_frames=2, start_frame=0, end_frame=None, verbose=False
):
    """
    slices a video into frames, returning a numpy array of frames
    :param src: path to video
    :param every_n_frames: stride between selected frames
    :param start_frame: first frame to take (inclusive)
    :param end_frame: last frame to take (inclusive)
    :return: (n x h x w x 3) tensor of sliced video
    """
    src = Path(src)
    if not src.exists():
        raise FileNotFoundError("Video file not found: {}".format(src))
    info = get_video_info(src)
    h, w, fc = info["height"], info["width"], info["frame_count"]
    if end_frame is None:
        end_frame = fc + 1
    out, _ = (
        ffmpeg.input(str(src))
        .trim(start_frame=start_frame, end_frame=end_frame + 1)
        .filter("select", "not(mod(n,{}))".format(every_n_frames))
        .output("pipe:", format="rawvideo", pix_fmt="rgb24", fps_mode="passthrough")
        .run(capture_stdout=True, quiet=not verbose)
    )
    video = np.frombuffer(out, np.uint8).reshape([-1, h, w, 3])
    return video


def get_frame_from_video(src, frame_index, verbose=False):
    """
    gets a single frame from a video
    :param src: path to video
    :param frame_num: frame number to get
    :return: (h x w x 3) tensor of frame
    """
    info = get_video_info(src)
    out, _ = (
        ffmpeg.input(str(src))
        .filter("select", "eq(n,{})".format(frame_index))
        .output("pipe:", format="rawvideo", pix_fmt="rgb24", fps_mode="passthrough")
        .run(capture_stdout=True, quiet=not verbose)
    )
    frame = np.frombuffer(out, np.uint8).reshape([info["height"], info["width"], 3])
    return frame


def trim_video(src, dst, start_frame=0, end_frame=None, verbose=False):
    """
    trim a video from disk and save it to disk
    note: will reencode video
    :param src: path to video
    :param dst: path to save trimmed video to
    :param start_frame: first frame to take (inclusive)
    :param end_frame: last frame to take (inclusive)
    :param verbose: if True, print ffmpeg output
    """
    src = Path(src)
    if not src.exists():
        raise FileNotFoundError("Video file not found: {}".format(src))
    info = get_video_info(src)
    if end_frame is None:
        end_frame = info["frame_count"] + 1
    out, _ = (
        ffmpeg.input(str(src))
        .trim(start_frame=start_frame, end_frame=end_frame + 1)
        .output(
            str(dst), vcodec="libx264", b=info["bit_rate"], pix_fmt=info["pixel_format"]
        )
        .overwrite_output()
        .run(capture_stdout=True, quiet=not verbose)
    )


class VideoReader:
    """
    A (very) basic video iterator
    """

    def __init__(
        self,
        video_path,
        h=None,
        w=None,
        every_n_frames=1,
        start_frame=0,
        end_frame=None,
        verbose=False,
    ):
        """
        :param video_path: path to video
        :param target_resolution: (h, w) tuple of target resolution (must have common divisor with original resolution)
        """
        self.video_path = Path(video_path)
        info = get_video_info(video_path)
        self.h, self.w, self.fps, self.fc = (
            info["width"],
            info["height"],
            info["fps"],
            info["frame_count"],
        )
        self.th = None
        self.tw = None
        if h is not None:
            if w is not None:
                self.th, self.tw = h, w
            else:
                raise ValueError("if h or w are specified, both must be specified")
        self.start_frame = start_frame
        if end_frame is None:
            self.end_frame = self.fc + 1
        else:
            self.end_frame = end_frame
        self.every_n_frames = every_n_frames
        self.stream = (
            ffmpeg.input(str(self.video_path))
            .trim(start_frame=self.start_frame, end_frame=self.end_frame)
            .filter("select", "not(mod(n,{}))".format(self.every_n_frames))
            .output("pipe:", format="rawvideo", pix_fmt="rgb24", fps_mode="passthrough")
            .run_async(pipe_stdout=True, quiet=not verbose)
        )

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self):
        if self.n < self.fc:
            in_bytes = self.stream.stdout.read(self.h * self.w * 3)
            if not in_bytes:  # should never happen
                self.stream.stdout.close()
                raise StopIteration
            frame = np.frombuffer(in_bytes, np.uint8).reshape([self.h, self.w, 3])
            if self.th is not None:
                if self.th != self.h or self.tw != self.w:
                    frame = cv2.resize(frame, (self.tw, self.th))
            self.n += 1
            return frame
        else:
            in_bytes = self.stream.stdout.read(self.h * self.w * 3)
            assert not in_bytes  # sanity check, no bytes should be left
            self.stream.stdout.close()
            raise StopIteration



import subprocess
import collections
import numpy as np
from pathlib import Path
import json
import time


def get_ffmpeg_version(verbose=False):
    """
    :return: ffmpeg version
    """
    try:
        ffmpeg_output = subprocess.run(
            [str(Path("ffmpeg")), "-version"],
            capture_output=True,
            text=True,
        ).stdout
        parts = ffmpeg_output.split()
        version = parts[parts.index("version") + 1]
    except ValueError:
        if verbose:
            print("gsoup warning: could not detect a valid ffmpeg version.")
    except FileNotFoundError:
        if verbose:
            print("gsoup warning: ffmpeg not found.")
    return version


get_ffmpeg_version(True)


class FPS:
    """
    calculates current fps and returns it, see https://stackoverflow.com/a/54539292
    example usage:
        fps = FPS()
        while True:
            # do something
            print(fps())
    """

    def __init__(self, avarageof=50):
        self.frametimestamps = collections.deque(maxlen=avarageof)

    def __call__(self):
        self.frametimestamps.append(time.time())
        nominator = len(self.frametimestamps)
        denominator = self.frametimestamps[-1] - self.frametimestamps[0]
        if (len(self.frametimestamps) > 1) and (denominator != 0):
            return nominator / denominator
        else:
            return 0.0


def probe_video(video_path):
    """
    Probe the video using ffprobe to extract useful metadata.
    Returns a dict with keys: width, height, frame_rate, codec, pix_fmt, bit_rate, nb_frames, duration.
    """
    video_path = Path(video_path)
    cmd = [
        "ffprobe",
        "-v",
        "error",
        "-select_streams",
        "v:0",
        "-show_entries",
        "stream=width,height,r_frame_rate,codec_name,pix_fmt,bit_rate,nb_frames,duration",
        "-of",
        "json",
        str(video_path),
    ]
    result = subprocess.run(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
    )
    info = json.loads(result.stdout)
    stream = info["streams"][0]
    try:
        num, den = stream["r_frame_rate"].split("/")
        frame_rate = float(num) / float(den)
    except Exception:
        frame_rate = None
    width = stream.get("width")
    height = stream.get("height")
    codec_name = stream.get("codec_name")
    pix_fmt = stream.get("pix_fmt")
    bit_rate = stream.get("bit_rate")
    nb_frames = stream.get("nb_frames")
    duration = stream.get("duration")
    if nb_frames is None and duration is not None and frame_rate is not None:
        nb_frames = int(float(duration) * frame_rate)
    else:
        try:
            nb_frames = int(nb_frames)
        except Exception:
            nb_frames = None
    return {
        "width": width,
        "height": height,
        "frame_rate": frame_rate,
        "codec": codec_name,
        "pix_fmt": pix_fmt,
        "bit_rate": bit_rate,
        "nb_frames": nb_frames,
        "duration": duration,
    }


def get_frame_timestamps(video_path):
    """
    Get a list of timestamps (in seconds) for each frame in the video.
    """
    video_path = Path(video_path)
    cmd = [
        "ffprobe",
        "-v",
        "error",
        "-select_streams",
        "v:0",
        "-show_entries",
        "frame=pkt_pts_time",
        "-of",
        "csv=p=0",
        str(video_path),
    ]
    result = subprocess.run(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
    )
    lines = result.stdout.strip().splitlines()
    timestamps = []
    for line in lines:
        try:
            timestamps.append(float(line))
        except ValueError:
            continue
    return timestamps


def load_video(video_path):
    video_path = Path(video_path)
    info = probe_video(video_path)
    width = info["width"]
    height = info["height"]
    if width is None or height is None:
        raise ValueError("Could not determine video dimensions.")
    frame_size = width * height * 3  # for rgb24
    cmd = [
        "ffmpeg",
        "-i",
        str(video_path),
        "-f",
        "rawvideo",
        "-pix_fmt",
        "rgb24",
        "-vcodec",
        "rawvideo",
        "-",
    ]
    pipe = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    raw_video = pipe.stdout.read()
    pipe.stdout.close()
    pipe.wait()
    n_frames = len(raw_video) // frame_size
    video = np.frombuffer(raw_video, dtype=np.uint8).reshape(
        (n_frames, height, width, 3)
    )
    return video


def save_video(
    frames,
    output_path,
    fps,
    bitrate=None,
    codec="libx264",
    pixel_format="yuv420p",
):
    """
    Write video frames to disk.
    Parameters:
      frames: (n_frames, height, width, 3) np array or an iterable yielding such frames.
      output_path: destination video file.
      fps: frames per second.
      bitrate: optional bitrate (e.g., "500k").
      codec: video codec to use.
      pixel_format: output pixel format.
    """
    output_path = Path(output_path)
    # Determine frame dimensions
    if isinstance(frames, np.ndarray):
        height, width = frames.shape[1:3]
        frame_iter = frames
    else:
        frame_iter = iter(frames)
        first_frame = next(frame_iter)
        height, width = first_frame.shape[:2]
        frame_iter = (frame for frame in [first_frame] + list(frame_iter))
    cmd = [
        "ffmpeg",
        "-y",
        "-f",
        "rawvideo",
        "-vcodec",
        "rawvideo",
        "-pix_fmt",
        "rgb24",
        "-s",
        f"{width}x{height}",
        "-r",
        str(fps),
        "-i",
        "-",
        "-an",
        "-vcodec",
        codec,
        "-pix_fmt",
        pixel_format,
    ]
    if bitrate:
        cmd.extend(["-b:v", str(bitrate)])
    cmd.append(str(output_path))
    pipe = subprocess.Popen(cmd, stdin=subprocess.PIPE)
    for frame in frame_iter:
        pipe.stdin.write(frame.astype(np.uint8).tobytes())
    pipe.stdin.close()
    pipe.wait()


def reverse_video(
    input_path, output_path, bitrate=None, codec="libx264", pixel_format="yuv420p"
):
    """
    Reverse a video on disk.
    """
    input_path = Path(input_path)
    output_path = Path(output_path)
    cmd = [
        "ffmpeg",
        "-y",
        "-i",
        str(input_path),
        "-vf",
        "reverse",
        "-an",
        "-vcodec",
        codec,
        "-pix_fmt",
        pixel_format,
    ]
    if bitrate:
        cmd.extend(["-b:v", str(bitrate)])
    cmd.append(str(output_path))
    subprocess.run(cmd, check=True)


def compress_video(
    input_path, output_path, bitrate, codec="libx264", pixel_format="yuv420p"
):
    """
    Compress a video by re-encoding it at the specified bitrate.
    """
    input_path = Path(input_path)
    output_path = Path(output_path)
    cmd = [
        "ffmpeg",
        "-y",
        "-i",
        str(input_path),
        "-an",
        "-vcodec",
        codec,
        "-pix_fmt",
        pixel_format,
        "-b:v",
        str(bitrate),
        str(output_path),
    ]
    subprocess.run(cmd, check=True)


def slice_video(
    input_path,
    start_frame,
    stop_frame,
    stride=1,
    fps=None,
    codec="libx264",
    pixel_format="yuv420p",
    bitrate=None,
    to_numpy=True,
    output_path=None,
):
    """
    Slice a video by selecting frames between start_frame and stop_frame with a given stride.
    Outputs a new video file.
    """
    input_path = Path(input_path)
    # Build a filter chain: select frames in range and skip by stride.
    vf_expr = (
        f"select='between(n\\,{start_frame}\\,{stop_frame})*"
        f"not(mod(n-{start_frame}\\,{stride}))',setpts=N/FRAME_RATE/TB"
    )
    base_cmd = ["ffmpeg", "-y", "-i", str(input_path), "-vf", vf_expr, "-an"]
    if fps:
        base_cmd.extend(["-r", str(fps)])
    if bitrate:
        base_cmd.extend(["-b:v", str(bitrate)])

    if to_numpy:
        # We want raw video output so we can convert to a numpy array.
        base_cmd.extend(
            ["-f", "rawvideo", "-pix_fmt", "rgb24", "-vcodec", "rawvideo", "pipe:1"]
        )
        proc = subprocess.Popen(
            base_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        info = probe_video(input_path)
        width = info["width"]
        height = info["height"]
        if width is None or height is None:
            raise ValueError("Could not determine video dimensions for slicing.")
        frame_size = width * height * 3
        raw_video = proc.stdout.read()
        proc.stdout.close()
        proc.wait()
        n_frames = len(raw_video) // frame_size
        video_array = np.frombuffer(raw_video, dtype=np.uint8).reshape(
            (n_frames, height, width, 3)
        )
        return video_array
    else:
        if output_path is None:
            raise ValueError("output_path must be provided when to_numpy is False.")
        output_path = Path(output_path)
        base_cmd.append(str(output_path))
        subprocess.run(base_cmd, check=True)


def get_single_frame(video_path, frame_index):
    """
    Extract a single frame from a video as a numpy array.
    """
    video_path = Path(video_path)
    info = probe_video(video_path)
    width = info["width"]
    height = info["height"]
    if width is None or height is None or info["frame_rate"] is None:
        raise ValueError("Insufficient video info to extract frame.")
    frame_size = width * height * 3
    fps = info["frame_rate"]
    # Calculate timestamp (in seconds) for the given frame index.
    timestamp = frame_index / fps
    cmd = [
        "ffmpeg",
        "-ss",
        str(timestamp),
        "-i",
        str(video_path),
        "-frames:v",
        "1",
        "-f",
        "rawvideo",
        "-pix_fmt",
        "rgb24",
        "pipe:1",
    ]
    result = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    raw_frame = result.stdout
    if len(raw_frame) < frame_size:
        raise ValueError("Could not extract frame.")
    frame = np.frombuffer(raw_frame, dtype=np.uint8).reshape((height, width, 3))
    return frame


def trim_video(
    input_path,
    output_path,
    start_frame,
    end_frame,
    bitrate=None,
    codec="libx264",
    pixel_format="yuv420p",
):
    """
    Trim a video using frame indices.
    """
    input_path = Path(input_path)
    output_path = Path(output_path)
    info = probe_video(input_path)
    if info["frame_rate"] is None:
        raise ValueError("Frame rate information unavailable.")
    fps = info["frame_rate"]
    start_time = start_frame / fps
    duration = (end_frame - start_frame) / fps
    cmd = [
        "ffmpeg",
        "-y",
        "-ss",
        str(start_time),
        "-i",
        str(input_path),
        "-t",
        str(duration),
        "-an",
        "-vcodec",
        codec,
        "-pix_fmt",
        pixel_format,
    ]
    if bitrate:
        cmd.extend(["-b:v", str(bitrate)])
    cmd.append(str(output_path))
    subprocess.run(cmd, check=True)


class VideoReader:
    def __init__(self, filename):
        self.filename = filename
        info = probe_video(filename)
        self.width = info["width"]
        self.height = info["height"]
        self.channels = 3  # For 'rgb24', there are 3 channels (R, G, B)
        # Launch ffmpeg to decode video into raw video frames
        self.proc = subprocess.Popen(
            ["ffmpeg", "-i", self.filename, "-f", "rawvideo", "-pix_fmt", "rgb24", "-"],
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,  # or subprocess.PIPE to capture errors
        )

    def __iter__(self):
        return self

    def __next__(self):
        frame_size = self.width * self.height * self.channels
        # Read enough bytes for one frame
        raw_frame = self.proc.stdout.read(frame_size)
        if len(raw_frame) < frame_size:
            # Clean up if we're done
            self.proc.stdout.close()
            self.proc.wait()
            raise StopIteration
        # Convert bytes to a numpy array and reshape to (height, width, channels)
        frame = np.frombuffer(raw_frame, dtype=np.uint8)
        return frame.reshape((self.height, self.width, self.channels))



import numpy as np
from .transforms import compose_rt, invert_rigid
from .rasterize import project_points, should_cull_tri
from .core import to_34, to_44, to_hom
from .geometry_basic import project_point_to_segment
import cv2
from scipy.optimize import least_squares


class KalmanFilter(object):
    # x0 - initial guess of the state vector
    # P0 - initial guess of the covariance matrix of the state estimation error
    # A,B,C - system matrices describing the system model
    # Q - covariance matrix of the process noise
    # R - covariance matrix of the measurement noise
    def __init__(self, x0, P0, A, B, C, Q, R):
        # initialize vectors and matrices
        self.x0 = x0
        self.P0 = P0
        self.A = A
        self.B = B
        self.C = C
        self.Q = Q
        self.R = R
        ####### members for debuggin #######
        self.currentTimeStep = 0
        self.estimates_aposteriori = []
        self.estimates_aposteriori.append(x0)
        self.estimates_apriori = []
        self.estimationErrorCovarianceMatricesAposteriori = []
        self.estimationErrorCovarianceMatricesAposteriori.append(P0)
        self.estimationErrorCovarianceMatricesApriori = []
        self.gainMatrices = []
        self.errors = []

    def set_matrices(self, A, B, C, Q, R):
        self.A = A
        self.B = B
        self.C = C
        self.Q = Q
        self.R = R

    # propagate x_{k-1}^{+} to compute x_{k}^{-}
    # this function also propagates P_{k-1}^{+} to compute P_{k}^{-}
    def predict(self, input_value, prev_xk_plus, prev_pk_plus):
        xk_minus = self.A * prev_xk_plus + self.B * input_value
        pk_minus = self.A * prev_pk_plus * (self.A.T) + self.Q
        return xk_minus, pk_minus

    # predict for k steps ahead (usefull for reducing latency in real systems)
    def predict_multi(self, inputValue, steps, prev_xk_plus):
        xk_minus = self.A * prev_xk_plus + self.B * inputValue
        for _ in range(steps - 1):
            xk_minus = self.A * xk_minus + self.B * inputValue
        return xk_minus

    # for debugging purposes, perdicts and stores the states as members of the class
    def predict_and_store(self, input_value):
        xk_minus = (
            self.A * self.estimates_aposteriori[self.currentTimeStep]
            + self.B * input_value
        )
        pk_minus = (
            self.A
            * self.estimationErrorCovarianceMatricesAposteriori[self.currentTimeStep]
            * (self.A.T)
            + self.Q
        )

        self.estimates_apriori.append(xk_minus)
        self.estimationErrorCovarianceMatricesApriori.append(pk_minus)

        self.currentTimeStep = self.currentTimeStep + 1

    # given a new measurement, perform the update state to fuse the measurement with the prediction
    def update(self, current_measurement, xk_minus, pk_minus):
        # gain matrix
        Kk = (
            pk_minus
            * (self.C.T)
            * np.linalg.inv(self.R + self.C * pk_minus * (self.C.T))
        )
        # prediction error
        error_k = current_measurement - self.C * xk_minus
        # a posteriori estimate
        xk_plus = xk_minus + Kk * error_k
        # a posteriori matrix update
        IminusKkC = np.matrix(np.eye(self.x0.shape[0])) - Kk * self.C
        pk_plus = IminusKkC * pk_minus * (IminusKkC.T) + Kk * (self.R) * (Kk.T)
        # return updated state
        return xk_plus, pk_plus

    # perform the update step and also stores the states as members of the class
    def update_and_store(self, current_measurement):
        Kk = (
            self.estimationErrorCovarianceMatricesApriori[self.currentTimeStep - 1]
            * (self.C.T)
            * np.linalg.inv(
                self.R
                + self.C
                * self.estimationErrorCovarianceMatricesApriori[
                    self.currentTimeStep - 1
                ]
                * (self.C.T)
            )
        )

        # prediction error
        error_k = (
            current_measurement
            - self.C * self.estimates_apriori[self.currentTimeStep - 1]
        )
        # a posteriori estimate
        xk_plus = self.estimates_apriori[self.currentTimeStep - 1] + Kk * error_k

        # a posteriori matrix update
        IminusKkC = np.matrix(np.eye(self.x0.shape[0])) - Kk * self.C
        Pk_plus = IminusKkC * self.estimationErrorCovarianceMatricesApriori[
            self.currentTimeStep - 1
        ] * (IminusKkC.T) + Kk * (self.R) * (Kk.T)

        # update the lists that store the vectors and matrices
        self.gainMatrices.append(Kk)
        self.errors.append(error_k)
        self.estimates_aposteriori.append(xk_plus)
        self.estimationErrorCovarianceMatricesAposteriori.append(Pk_plus)


class HullTracker:
    def __init__(self, init_pose, params):
        self.params = params
        self.dist_coeff = np.zeros((5, 1)).astype(np.float32)
        # define the initial camera pose (using the gt object pose).
        # o2c = w2c @ o2w
        o2c = to_44(params["w2c"]) @ to_44(init_pose)
        # initial_rmat = np.eye(3).astype(np.float32)  # rand_rot_mat[0]
        o2c = to_34(o2c)
        initial_rmat = o2c[:3, :3]
        initial_rvec, _ = cv2.Rodrigues(initial_rmat)
        # initial_tvec = np.zeros(3).astype(np.float32)  # rand_rot_trans[0]
        initial_tvec = o2c[:, -1]
        # Assume an initial pose is given [rvec_x, rvec_y, rvec_z, tvec_x, tvec_y, tvec_z]
        self.poses = []
        self.corres = []
        self.cur_pose = np.concatenate((initial_rvec[:, 0], initial_tvec), axis=-1)

    def get_convex_hull_ideal(self, V, K, w2c):
        projected = project_points(
            self.params["v"], self.params["K"], self.params["w2c"]
        )[:, :2]
        hull = cv2.convexHull(projected)
        return hull

    def get_convex_hull(self, frame):
        # todo
        x = cv2.findContours(frame)
        hull = cv2.convexHull(x)
        return hull

    def extract_silhouette(self, projected_points):
        """
        For simplicity, we use the convex hull as the silhouette.
        Returns:
        silhouette: (M,2) array of 2D points on the hull.
        hull_indices: indices into the projected_points array.
        """
        pts = np.array(projected_points, dtype=np.float32).reshape(-1, 1, 2)
        hull_indices = cv2.convexHull(pts, returnPoints=False)
        hull_indices = hull_indices.flatten()
        silhouette = pts[hull_indices].reshape(-1, 2)
        return silhouette, hull_indices

    def find_closest_point(self, p, observed_silhouette):
        """
        For a given 2D point p, find the closest point on the observed silhouette.
        Returns:
        best_point: the closest point on an edge,
        best_tangent: the unit tangent of that edge.
        """
        best_distance = float("inf")
        best_point = None
        best_tangent = None
        N = observed_silhouette.shape[0]
        for i in range(N):
            a = observed_silhouette[i]
            b = observed_silhouette[(i + 1) % N]
            candidate = project_point_to_segment(p, a, b)
            d = np.linalg.norm(p - candidate)
            if d < best_distance:
                best_distance = d
                best_point = candidate
                tangent = b - a
                norm_t = np.linalg.norm(tangent)
                best_tangent = tangent / norm_t if norm_t > 0 else np.array([0, 0])
        return best_point, best_tangent

    def track(self, frame, cur_v, **kwargs):
        observed_silhouette = self.get_convex_hull_ideal(
            cur_v, params["K"], params["w2c"]
        )
        # observed_silhouette = self.get_convex_hull(frame)
        for iter in range(self.params["iters_per_frame"]):
            # convert current o2c to o2w
            rvec = self.cur_pose[0:3]
            tvec = self.cur_pose[3:6]
            rmat, _ = cv2.Rodrigues(rvec)
            o2c = compose_rt(rmat[None, ...], tvec[None, :, 0], square=True)[0]
            c2w = invert_rigid(to_44(self.params["w2c"])[None, ...], square=True)[0]
            o2w = to_34(c2w @ o2c)
            # apply o2w to object
            V_test = (o2w @ to_hom(self.params["v"]).T).T
            projected_points = project_points(
                V_test, self.params["K"], self.params["w2c"]
            )[:, :2]
            proj_silhouette, silhouette_indices = self.extract_silhouette(
                projected_points
            )
            correspondences = []
            for p in proj_silhouette:
                q, tangent = self.find_closest_point(p, observed_silhouette)
                correspondences.append((p, (q, tangent)))
            target_points = np.array([c[1][0] for c in correspondences])
            mean_error = np.mean(
                np.linalg.norm(proj_silhouette - target_points, axis=1)
            )
            print(mean_error)
            success, rvec, tvec = cv2.solvePnP(
                V[silhouette_indices],  # should be filtered to only use correspondences
                target_points,
                self.params["K"],
                self.dist_coeff,
                self.cur_pose[0:3],
                self.cur_pose[3:6],
                useExtrinsicGuess=True,
                flags=cv2.SOLVEPNP_ITERATIVE,
            )  # finds best fitting w2c
            # breakpoint()
            if not success:
                raise RuntimeError("solvePnP failed during refinement.")
            self.cur_pose = np.concatenate((rvec[:, 0], tvec), axis=-1)
            self.poses.append(self.cur_pose)
            self.corres.append(correspondences)

    def get_results():
        o2ws = []
        for pose in self.poses:
            rvec = pose[0:3]
            tvec = pose[3:6]
            rmat, _ = cv2.Rodrigues(rvec)
            o2c = compose_rt(rmat[None, ...], tvec[None, :, 0], square=True)[0]
            c2w = invert_rigid(to_44(self.params["w2c"])[None, ...], square=True)[0]
            o2w = to_34(c2w @ o2c)
            o2ws.append(o2w)
        return o2ws, self.corres


class NaiveEdgeTracker:
    def __init__(self, init_pose, params):
        self.params = params
        # define the initial camera pose (using the gt object pose).
        # o2c = w2c @ o2w
        o2c = to_44(params["w2c"]) @ to_44(init_pose)
        # initial_rmat = np.eye(3).astype(np.float32)  # rand_rot_mat[0]
        initial_rmat = o2c[:3, :3]
        initial_rvec, _ = cv2.Rodrigues(initial_rmat)
        # initial_tvec = np.zeros(3).astype(np.float32)  # rand_rot_trans[0]
        initial_tvec = o2c[:, -1]
        # Get our 3D model and sample points along its edges
        self.sampled_points = self.sample_edge_points(n_samples=10)
        # breakpoint()
        # Assume an initial pose is given [rvec_x, rvec_y, rvec_z, tvec_x, tvec_y, tvec_z]
        self.poses = []
        self.cur_pose = np.concatenate((initial_rvec[:, 0], initial_tvec), axis=-1)

    def sample_edge_points(self, n_samples=10):
        """
        Sample points uniformly along each edge of the model.
        """
        points = []
        cam_pose = invert_rigid(to_44(self.params["w2c"])[None, :])[
            0, :3, -1
        ]  # get cam pose
        for i, (a, b) in enumerate(self.params["e"]):
            v0, v1, v2, _ = self.params["v"][self.params["f"][self.params["e2f"][i][0]]]
            cond1 = should_cull_tri(v0, v1, v2, cam_pose)
            v0, v1, v2, _ = self.params["v"][self.params["f"][self.params["e2f"][i][1]]]
            cond2 = should_cull_tri(v0, v1, v2, cam_pose)
            if cond1 and cond2:
                continue
            p0, p1 = self.params["v"][a], self.params["v"][b]
            # Sample n_samples points (including endpoints)
            for t in np.linspace(0, 1, n_samples):
                pt = (1 - t) * p0 + t * p1
                points.append(pt)
        return np.array(points, dtype=np.float32)

    def track_project_points(self, points, rvec, tvec, K):
        rmat, _ = cv2.Rodrigues(rvec)
        w2c = compose_rt(rmat[None, ...], tvec[None, ...])
        projected_points = project_points(points, K, w2c)[:, :2]
        return projected_points

    def cost_function(self, params, sampled_points, dist_transform, K):
        """
        Cost function that returns the distance from each projected sample point
        to the nearest image edge (via the distance transform).
        params: 6-vector (3 for rotation [Rodrigues], 3 for translation)
        """
        rvec = params[0:3]
        tvec = params[3:6]
        projected_pts = self.track_project_points(sampled_points, rvec, tvec, K)[:, :2]
        cost = []
        for pt in projected_pts:
            x, y = np.round(pt).astype(np.uint32)
            # Penalize if projected point is outside the image bounds
            if (
                x < 0
                or x >= dist_transform.shape[1]
                or y < 0
                or y >= dist_transform.shape[0]
            ):
                cost.append(np.array([100.0]))
            else:
                cost.append(dist_transform[y, x])
        return np.array(cost).squeeze()

    def track(self, frame):
        # Preprocess the frame: grayscale and edge detection
        # gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        # edges_img = cv2.Canny(gray, 50, 150)
        # Invert edge image so that edges become zeros and background is nonzero.
        # _, inv_edges_img = cv2.threshold(edges_img, 50, 255, cv2.THRESH_BINARY_INV)
        inv_edges_img = 255 - ((frame > 0) * 255).astype(np.uint8)
        # Then compute the distance transform (distance from each pixel to the nearest edge).
        dist_transform = cv2.distanceTransform(inv_edges_img, cv2.DIST_L2, 5)
        # Optimize the pose parameters to align the model's projected edges with the image edges.
        for ii in range(self.params["iters_per_frame"]):
            res = least_squares(
                self.cost_function,
                self.cur_pose,
                args=(self.sampled_points, dist_transform, self.params["K"]),
                verbose=0,
            )
            self.cur_pose = res.x  # update pose for next iteration
            self.poses.append(self.cur_pose)

    def get_results(self):
        object_poses = []
        for i in range(len(self.poses)):
            rvec = self.poses[i][0:3]
            rmat, _ = cv2.Rodrigues(rvec)
            tvec = self.poses[i][3:6]
            # breakpoint()
            o2c = compose_rt(rmat[None, ...], tvec[None, ...], square=True)[0]
            # c2o = gsoup.invert_rigid(o2c) # c2o = w2c @ o2w
            c2w = invert_rigid(to_44(self.params["w2c"])[None, ...], square=True)[0]
            o2w = to_34(c2w @ o2c)
            object_poses.append(o2w)
        return np.array(object_poses), None



import numpy as np
from .core import (
    to_44,
    compose_rt,
    normalize,
    broadcast_batch,
    is_np,
    map_range,
    to_torch,
)
import torch


def ndc_to_screen(ndc, width, height):
    """
    convert 2D NDC (normalized device coordiantes) to screen space
    :param ndc: the b x 2 coordinates
    :param width: the width of the screen in pixels
    :param height: the height of the screen in pixels
    """
    size = np.array([width, height], dtype=np.float32)[None, :]
    ndc_norm = map_range(ndc, -1.0, 1.0, 0.0, 1.0)
    if is_np(ndc):
        screen = (ndc_norm * size).astype(np.int32)
    else:
        screen = (
            ndc_norm * to_torch(size, device=ndc_norm.device, dtype=torch.float32)
        ).type(torch.uint32)
    return screen


def screen_to_ndc(screen, width, height):
    """
    convert screen space coordinates to NDC
    :param ndc: the b x 2 coordinates
    :param width: the width of the screen in pixels
    :param height: the height of the screen in pixels
    """
    size = np.array([width - 1, height - 1], dtype=np.float32)[None, :]
    if not is_np(screen):
        size = to_torch(size, device=screen.device)
    ndc = map_range(screen, 0.0, size, -1.0, 1.0)
    return ndc


def sincos(a, degrees=True):
    """
    sin and cos of an angle
    :param a: angle
    :param degrees: if True, a is in degrees
    """
    if degrees:
        a = np.deg2rad(a)
    return np.sin(a), np.cos(a)


def translate(t):
    """
    creates a translation matrix from a translation vector
    :param t: translation vector or batch of translation vectors
    :return: 4x4 translation matrix
    """
    if t.shape[-1] != 3:
        raise ValueError("translation vector must be 3d")
    if t.ndim == 1:
        t = t[None, :]
    mat = np.concatenate((np.eye(3)[None, :, :], t[:, :, None]), axis=-1)
    return to_44(mat)


def scale(s):
    """
    creates a scaling matrix from a scaling vector
    :param s: scaling vector or batch of scaling vectors
    :return: nx4x4 scaling matrix
    """
    if s.shape[-1] != 3:
        raise ValueError("translation vector must be 3d")
    if s.ndim == 1:
        s = s[None, :]
    mat = np.diag(s)
    return to_44(mat)


def rotate(a, r):
    """
    creates a rotation matrix from an angle a and an axis of rotation r
    """
    s, c = sincos(a)
    r = normalize(r)
    nc = 1 - c
    x, y, z = r
    return np.array(
        [
            [x * x * nc + c, x * y * nc - z * s, x * z * nc + y * s, 0],
            [y * x * nc + z * s, y * y * nc + c, y * z * nc - x * s, 0],
            [x * z * nc - y * s, y * z * nc + x * s, z * z * nc + c, 0],
            [0, 0, 0, 1],
        ]
    )


def rotx(a, degrees=True):
    """
    creates a homogeneous 3D rotation matrix around the x axis
    a: angle
    degrees: if True, a is in degrees, else radians
    """
    s, c = sincos(a, degrees)
    return np.array([[1, 0, 0, 0], [0, c, -s, 0], [0, s, c, 0], [0, 0, 0, 1]])


def roty(a, degrees=True):
    """
    creates a homogeneous 3D rotation matrix around the y axis
    a: angle
    degrees: if True, a is in degrees, else radians
    """
    s, c = sincos(a, degrees)
    return np.array([[c, 0, s, 0], [0, 1, 0, 0], [-s, 0, c, 0], [0, 0, 0, 1]])


def rotz(a, degrees=True):
    """
    creates a homogeneous 3D rotation matrix around the z axis
    a: angle
    degrees: if True, a is in degrees, else radians
    """
    s, c = sincos(a, degrees)
    return np.array([[c, -s, 0, 0], [s, c, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])


def find_rigid_transform(A, B):
    """
    finds best 3d rigid transformation between pc a and pc b (in the least squares sense) based on "Least-Squares Rigid Motion Using SVD"
    :param A: point cloud a
    :param B: point cloud b
    :return: W such that W@A = B
    """
    centroid_A = np.mean(A, axis=0)
    centroid_B = np.mean(B, axis=0)

    A_mean = A - centroid_A
    B_mean = B - centroid_B

    H = A_mean.T @ B_mean
    U, S, Vt = np.linalg.svd(H)

    flip = np.linalg.det(Vt.T @ U.T)
    ones = np.identity(len(Vt))
    ones[-1, -1] = flip
    R = Vt.T @ ones @ U.T
    t = centroid_B - R @ centroid_A
    return compose_rt(R[None, :], t[None, :], square=True)


def find_affine_transformation(A, B):
    """
    finds best 3d affine transformation between pc a and pc b (in the least squares sense)
    :param A: point cloud a
    :param B: point cloud b
    :return: W such that W@A = B
    """
    # assemble A matrix for least squares
    assert len(A) >= 4
    new_A = np.c_[A, np.ones(len(A))]
    new_A = np.repeat(np.tile(new_A, (1, 3)), 3, axis=0)
    # now A is 3nx12
    # zero columns depending on index of point
    new_A[::3, 4:] = 0
    new_A[1::3, :4] = 0
    new_A[1::3, 8:] = 0
    new_A[2::3, :8] = 0
    new_B = B.flatten()
    if len(A) == 4:
        # solve linear system
        W = np.linalg.inv(new_A) @ new_B
    else:
        W = np.linalg.lstsq(new_A, new_B, rcond=None)[0]
    return to_44(W.reshape(3, 4))


def decompose_affine(A44):
    """
    Decompose 4x4 homogenous affine matrix into parts.
    The parts are translations, rotations, zooms, shears.
    Returns
    -------
    T : array, shape (3,)
       Translation vector
    R : array shape (3,3)
        rotation matrix
    Z : array, shape (3,)
       scale vector.  May have one negative zoom to prevent need for negative
       determinant R matrix above
    S : array, shape (3,)
       Shear vector, such that shears fill upper triangle above
       diagonal to form shear matrix (type ``striu``).
    Implementation of "Decomposing a matrix into simple transformations".
    """
    A44 = np.asarray(A44)
    T = A44[:-1, -1]
    RZS = A44[:-1, :-1]
    # compute scales and shears
    M0, M1, M2 = np.array(RZS).T
    # extract x scale and normalize
    sx = np.sqrt(np.sum(M0**2))
    M0 /= sx
    # orthogonalize M1 with respect to M0
    sx_sxy = np.dot(M0, M1)
    M1 -= sx_sxy * M0
    # extract y scale and normalize
    sy = np.sqrt(np.sum(M1**2))
    M1 /= sy
    sxy = sx_sxy / sx
    # orthogonalize M2 with respect to M0 and M1
    sx_sxz = np.dot(M0, M2)
    sy_syz = np.dot(M1, M2)
    M2 -= sx_sxz * M0 + sy_syz * M1
    # extract z scale and normalize
    sz = np.sqrt(np.sum(M2**2))
    M2 /= sz
    sxz = sx_sxz / sx
    syz = sy_syz / sy
    # Reconstruct rotation matrix, ensure positive determinant
    Rmat = np.array([M0, M1, M2]).T
    if np.linalg.det(Rmat) < 0:
        sx *= -1
        Rmat[:, 0] *= -1
    return T, Rmat, np.array([sx, sy, sz]), np.array([sxy, sxz, syz])


def invert_rigid(rt, square=False):
    """
    inverts a rigid transformation (n, 4, 4) efficiently
    rt: (n, 4, 4) batch of rigid transformations
    square: if false, will return a (n, 3, 4) by removing the last row
    """
    if rt.ndim != 3:
        return ValueError("rt must have 3 dimensions")
    if rt.shape[1] != 4 or rt.shape[2] != 4:
        return ValueError("rt must have shape (n, 4, 4)")
    if is_np(rt):
        R = np.transpose(rt[:, :3, :3], axes=(0, 2, 1))
        t = (-R @ rt[:, :-1, -1].T)[:, :, 0]
        return compose_rt(R, t, square)
    else:
        return NotImplementedError


def look_at_np(eye, at, up, opengl=False):
    """
    returns a batch of look_at transforms 4x4 (camera->world, the inverse of a ModelView matrix)
    will broadcast upon batch dimension if necessary.
    :param eye: n x 3 from vectors in world space
    :param at: n x 3 at vector in world space
    :param up: n x 3 up vector in world space
    :param opengl: if True, output will be in OpenGL coordinates (z backward, y up) otherwise (z forward, y down)
    :return: n x 4 x 4 transformation matrices (camera->world, the inverse of a ModelView matrix)
    """
    eye, at, up = broadcast_batch(eye, at, up)
    forward = at - eye
    forward = forward / np.linalg.norm(forward, axis=-1, keepdims=True)
    right = np.cross(forward, up)
    right = right / np.linalg.norm(right, axis=-1, keepdims=True)
    up = np.cross(forward, right)
    up = up / np.linalg.norm(up, axis=-1, keepdims=True)
    rot = np.concatenate((right[..., None], up[..., None], forward[..., None]), axis=-1)
    c2w = compose_rt(rot, eye, square=True)
    if opengl:
        c2w = opencv_c2w_to_opengl_c2w(c2w)
    return c2w


def look_at_torch(
    eye: torch.Tensor,  # 3
    at: torch.Tensor,  # 3
    up: torch.Tensor,  # 3
    opengl: bool = False,
) -> torch.Tensor:  # 4,4
    """
    creates a lookat transform matrix
    :param eye: where the camera is
    :param at: where the camera is looking
    :param up: the up vector of world space
    :param opengl: if True, output will be in OpenGL coordinates (z backward, y up) otherwise (z forward, y down)
    :return: 4x4 lookat transform matrix
    """
    # todo batch support
    z = (at - eye).type(torch.float32)
    z /= torch.norm(z)
    x = torch.linalg.cross(z, up).type(torch.float32)
    x /= torch.norm(x)
    y = torch.linalg.cross(z, x).type(torch.float32)
    y /= torch.norm(y)
    c2w = torch.eye(4, device=z.device)
    c2w[:3, :3] = torch.stack([x, y, z], dim=1)
    c2w[:3, 3] = eye
    if opengl:
        c2w = opencv_c2w_to_opengl_c2w(c2w)
    return c2w


def orthographic_projection(l, r, b, t, n, f):
    """
    creates an orthographic projection matrix (OpenGL convention)
    :param l: left
    :param r: right
    :param b: bottom
    :param t: top
    :param n: near
    :param f: far
    :return: 4x4 orthographic projection matrix
    """
    dx = r - l
    dy = t - b
    dz = f - n
    rx = -(r + l) / (r - l)
    ry = -(t + b) / (t - b)
    rz = -(f + n) / (f - n)
    return np.array(
        [
            [2.0 / dx, 0, 0, rx],
            [0, 2.0 / dy, 0, ry],
            [0, 0, -2.0 / dz, rz],
            [0, 0, 0, 1],
        ]
    )


def perspective_projection(fovy=45, aspect=1.0, n=0.1, f=100.0):
    """
    creates a perspective projection matrix (OpenGL convention)
    :param fovy: field of view in y direction
    :param aspect: aspect ratio
    :param n: near plane
    :param f: far plane
    :return: 4x4 projection matrix
    """
    s = 1.0 / np.tan(np.deg2rad(fovy) / 2.0)
    sx, sy = s / aspect, s
    zz = (f + n) / (n - f)
    zw = 2 * f * n / (n - f)
    return np.array([[sx, 0, 0, 0], [0, sy, 0, 0], [0, 0, zz, zw], [0, 0, -1, 0]])


def frustum_projection(x0, x1, y0, y1, z0, z1):
    """
    creates a projection matrix from a frustum (OpenGL convention)
    :param x0: left
    :param x1: right
    :param y0: bottom
    :param y1: top
    :param z0: near
    :param z1: far
    :return: 4x4 projection matrix
    """
    a = (x1 + x0) / (x1 - x0)
    b = (y1 + y0) / (y1 - y0)
    c = -(z1 + z0) / (z1 - z0)
    d = -2 * z1 * z0 / (z1 - z0)
    sx = 2 * z0 / (x1 - x0)
    sy = 2 * z0 / (y1 - y0)
    return np.array([[sx, 0, a, 0], [0, sy, b, 0], [0, 0, c, d], [0, 0, -1, 0]])


def opengl_project_from_opencv_intrinsics(
    opencv_intrinsics, width, height, near=0.1, far=100.0, window_coords_inverted=False
):
    """
    given a matrix K from opencv, returns the corresponding projection matrix for OpenGL ("Eye/Camera/View space -> Clip Space")
    :param opencv_intrinsics: 3x3 intrinsics matrix from opencv
    :param width: width of the image
    :param height: height of the image
    :param near: near plane
    :param far: far plane
    :param window_coords_inverted: if True, the y axis is inverted (y down)
    :return: 4x4 projection matrix for OpenGL (note: column major)
    """
    fx = opencv_intrinsics[0, 0]
    fy = opencv_intrinsics[1, 1]
    cx = opencv_intrinsics[0, 2]
    cy = opencv_intrinsics[1, 2]
    factor = 1.0
    if window_coords_inverted:
        factor = -1.0
    opengl_mtx = np.array(
        [
            [2 * fx / width, 0.0, (width - 2 * cx) / width, 0.0],
            [0.0, factor * 2 * fy / height, factor * (height - 2 * cy) / height, 0.0],
            [0.0, 0.0, (-far - near) / (far - near), -2.0 * far * near / (far - near)],
            [0.0, 0.0, -1.0, 0.0],
        ]
    )
    return opengl_mtx


def opencv_intrinsics_from_opengl_project(opengl_project, width, height):
    """
    given a projection matrix from OpenGL ("Eye/Camera/View space -> Clip Space"), returns a matrix K for opencv
    :param opengl_project: 4x4 projection matrix from OpenGL
    :param width: width of the image
    :param height: height of the image
    :return: 3x3 intrinsics matrix for opencv
    note: assumes camera center is middle of image
    """
    fx = opengl_project[0, 0] * width / 2
    fy = opengl_project[1, 1] * height / 2
    cx = width / 2
    cy = height / 2
    opencv_mtx = np.array([[fx, 0.0, cx], [0.0, fy, cy], [0.0, 0.0, 1]])
    return opencv_mtx


def opengl_c2w_to_opencv_c2w(opengl_transforms, is_column_major=False):
    """
    given a modelview matrix (World space->Eye/Camera/View space) where z is backward and y is up,
    converts its coordinate system to opencv convention (z forward, y down)
    :param opengl_transforms: 4x4 modelview matrix or batch of 4x4 modelview matrices
    :param is_column_major: whether the input matrix is column major or row major
    :return: 4x4 modelview matrix in opencv convention
    """
    if opengl_transforms.ndim == 2:
        my_transforms = opengl_transforms[None, ...]
    else:
        my_transforms = opengl_transforms
    if my_transforms.shape[1:] != (4, 4):
        raise ValueError("transform must be 4x4 or batch of 4x4")
    if is_column_major:
        if is_np(my_transforms):
            my_transforms = my_transforms.transpose(0, 2, 1)
        else:
            my_transforms = my_transforms.permute(0, 2, 1)
        my_transforms[:, 1, :] *= -1
        my_transforms[:, 2, :] *= -1
    else:
        my_transforms[:, :, 1] *= -1
        my_transforms[:, :, 2] *= -1
    return my_transforms.reshape(opengl_transforms.shape)


def opencv_c2w_to_opengl_c2w(opencv_transform, to_column_major=False):
    """
    converts coordinates of "vision" (opencv) to OpenGL coordinates by flipping y and z axes
    """
    return opengl_c2w_to_opencv_c2w(opencv_transform, to_column_major)


def create_random_cameras_on_unit_sphere(
    n_cams, radius, normal=None, opengl=False, device="cpu"
):
    """
    creates a batch of world2view ("ModelView" matrix) and view2clip ("Projection" matrix) transforms on a unit sphere looking at the center
    :param n_cams: number of cameras
    :param radius: radius of the sphere
    :param normal: if provided, only the hemisphere in the direction of the normal is sampled
    :param opengl: if True, the coordinate system is converted to openGL convention (z backward, y up)
    :param device: device to put the tensors on
    :return: world2view (world2cam), view2clip (requires further processing if opengl=False)
    """
    locs = torch.randn((n_cams, 3), device=device)
    locs = torch.nn.functional.normalize(locs, dim=1, eps=1e-6)
    if normal is not None:
        if normal.ndim == 1:
            normal = normal[None, :]
        normal = torch.nn.functional.normalize(normal, dim=-1, eps=1e-6)
        dot_product = (locs[:, None, :] @ normal[:, :, None]).squeeze()
        locs[dot_product < 0] *= -1
    locs = locs * radius
    matrices = torch.empty((n_cams, 4, 4), dtype=torch.float32, device=device)
    for i in range(len(locs)):
        matrices[i] = look_at_torch(
            locs[i],
            torch.zeros(3, dtype=torch.float32, device=device),
            torch.tensor([0.0, 0.0, 1.0], device=device),
            opengl=opengl,
        )
    v2w = matrices  # c2w
    w2v = torch.inverse(v2w)
    v2c = perspective_projection()
    v2c = np.tile(v2c[None, :], (n_cams, 1, 1))
    v2c = torch.tensor(v2c, dtype=torch.float32, device=device)
    return w2v, v2c


def pixels_in_world_space(camera_wh, K, Rt):
    """
    computes pinhole camera pixel locations in world space
    :param camera_wh: resolution (width, height) as a tuple
    :param K: the camera instrinsics matrix 3x3
    :param Rt: the camera to world matrix 4x4
    :return: np array n x 3 for pixels locations (in world space)
    """
    x, y = np.meshgrid(
        np.arange(camera_wh[0], dtype=np.float32),
        np.arange(camera_wh[1], dtype=np.float32),
        indexing="xy",
    )
    x = x.reshape(-1)[:, None]
    y = y.reshape(-1)[:, None]
    ones = np.ones(len(x))[:, None]
    pixels_image = np.concatenate((x, y, ones), axis=-1)
    pixels_camera = np.linalg.inv(K) @ pixels_image.T
    pixels_camera_hom = np.concatenate(
        (pixels_camera.T, np.ones(len(pixels_camera.T))[:, None]), axis=-1
    )
    pixels_world = Rt @ pixels_camera_hom.T
    pc = pixels_world.T[:, :3]
    return pc


def random_qvec(n: int):
    """
    generate random quaternions representing rotations
    :param n: Number of quaternions in a batch to return.
    :return: Quaternions as tensor of shape (N, 4).
    """
    o = np.random.randn(n, 4)
    s = (o * o).sum(1)
    denom = np.copysign(np.sqrt(s), o[:, 0])[:, None]
    o = o / denom
    return o


def random_affine(ang_range=20.0, trans_range=10.0, scale=2.0):
    """
    generates a random 2D affine transformation matrix
    """
    R = np.eye(3)
    ang_rot = np.random.uniform(ang_range) - ang_range / 2
    tx = trans_range * np.random.uniform() - trans_range / 2
    ty = trans_range * np.random.uniform() - trans_range / 2
    sx = np.random.rand() * scale
    sy = np.random.rand() * scale
    R[0, 0] = np.cos(ang_rot * np.pi / 180)
    R[0, 1] = -np.sin(ang_rot * np.pi / 180)
    R[1, 0] = np.sin(ang_rot * np.pi / 180)
    R[1, 1] = np.cos(ang_rot * np.pi / 180)
    T = np.eye(3)
    T[0, 2] = tx
    T[1, 2] = ty
    S = np.eye(3)
    S[0, 0] = sx
    S[1, 1] = sy
    H = T @ S @ R
    return H


def random_perspective():
    """
    generates a random 2D perspective transformation matrix
    """
    P = random_affine()
    P[2, 0] = np.random.rand() / 100
    P[2, 1] = np.random.rand() / 100
    return P


def random_vectors_on_sphere(n, normal=None, device="cpu"):
    """
    create a batch of uniformly distributed random unit vectors on a sphere
    note: if normal is provided, returns random unit vectors on the hemisphere around the normal, but isn't uniform anymore.
    :param n: number of vectors
    :param normal: normals to orient the hemisphere (,3) or (n,3)
    :param device: device to put the tensors on
    :return: tensor of shape (n, 3)
    """
    locs = torch.randn((n, 3), device=device)
    locs = torch.nn.functional.normalize(locs, dim=1, eps=1e-6)
    if normal is not None:
        if normal.ndim == 1:
            normal = normal[None, :]
        normal = torch.nn.functional.normalize(normal, dim=-1, eps=1e-6)
        dot_product = (locs[:, None, :] @ normal[:, :, None]).squeeze()
        locs[dot_product < 0] *= -1
    return locs


def vec2skew(v):
    """
    returns the skew operator matrix given a vector
    :param v:  (3, ) torch tensor
    :return:   (3, 3)
    """
    return batch_vec2skew(v[None, :])[0]


def batch_vec2skew(v):
    """
    returns a batch of skew operator matrices given a batch of vectors
    :param v:  (B, 3) torch tensor
    :return:   (B, 3, 3)
    """
    zero = torch.zeros_like(v[:, 0:1])
    skew_v0 = torch.cat([zero, -v[:, 2:3], v[:, 1:2]], dim=1)  # (B, 3)
    skew_v1 = torch.cat([v[:, 2:3], zero, -v[:, 0:1]], dim=1)
    skew_v2 = torch.cat([-v[:, 1:2], v[:, 0:1], zero], dim=1)
    skew_v = torch.stack([skew_v0, skew_v1, skew_v2], dim=1)  # (B, 3, 3)
    return skew_v  # (B, 3, 3)


def batch_rotvec2mat(r: torch.Tensor):
    """so(3) vector to SO(3) matrix
    :param r: (3, ) axis-angle, torch tensor
    :return:  (3, 3)
    """
    if r.ndim != 2:
        raise ValueError("r must be 2D tensor.")
    skew_r = batch_vec2skew(r)  # (3, 3)
    norm_r = r.norm(dim=-1, keepdim=True)[:, :, None] + 1e-15
    eye = torch.eye(3, dtype=torch.float32, device=r.device)[None, :]
    R = (
        eye
        + (torch.sin(norm_r) / norm_r) * skew_r
        + ((1 - torch.cos(norm_r)) / norm_r**2) * (skew_r @ skew_r)
    )
    return R


def rotvec2mat(r: torch.Tensor):
    """so(3) vector to SO(3) matrix
    :param r: (3, ) axis-angle, torch tensor
    :return:  (3, 3)
    """
    return batch_rotvec2mat(r[None, :])[0]


def mat2rotvec(r: torch.Tensor):
    """SO(3) matrix to so(3) vector
    :param r: (3, ) axis-angle, torch tensor
    :return:  (3, 3)
    """
    e, v = torch.linalg.eig(r)
    rotvec = v.real[:, torch.isclose(e.real, torch.ones(1))].squeeze()
    up = torch.tensor([0, 1, 0], dtype=torch.float)
    # ortho = torch.cross(rotvec, up)
    # ortho = ortho / ortho.norm(keepdim=True)
    # angle1 = torch.arccos(ortho.dot(r @ ortho))
    angle2 = torch.arccos((torch.trace(r) - 1) / 2)
    raise NotImplementedError("please verify this function implementation before usage")
    return rotvec * angle2


def qvec2mat(qvecs):
    """
    Converts a batch of quaternions to a batch of rotation matrices
    :param qvec: tensor of shape (n, 4) where real part is first [a + bi + cj + dk] => [a, b, c, d]
    :return: rotation matrix (n, 3, 3)
    """
    if qvecs.ndim != 2:
        raise ValueError("qvec should have ndim == 2")
    if qvecs.shape[-1] != 4:
        raise ValueError("quaternions must be of shape (..., 4)")
    if is_np(qvecs):
        r, i, j, k = [x[0] for x in np.split(qvecs, 4, -1)]
        stacking_func = np.stack
    else:
        r, i, j, k = torch.unbind(qvecs, -1)
        stacking_func = torch.stack
    two_s = 2.0 / (qvecs * qvecs).sum(-1)
    o = stacking_func(
        [
            1 - two_s * (j * j + k * k),
            two_s * (i * j - k * r),
            two_s * (i * k + j * r),
            two_s * (i * j + k * r),
            1 - two_s * (i * i + k * k),
            two_s * (j * k - i * r),
            two_s * (i * k - j * r),
            two_s * (j * k + i * r),
            1 - two_s * (i * i + j * j),
        ],
        -1,
    )
    return o.reshape(qvecs.shape[:-1] + (3, 3))


def mat2qvec_numpy(R: np.array):
    """
    converts a rotation matrix to a quaternion
    :param R: np array of size 3x3
    :return: qvec (4,) xyzw
    """
    q = np.empty((R.shape[0], 4), dtype=R.dtype)
    trace = np.trace(R)
    expanded_trace = np.concatenate((np.diag(R), trace), axis=0)
    choice = np.argmax(expanded_trace, axis=-1)
    mask = choice != 3
    i = choice[mask]
    j = (i + 1) % 3
    k = (j + 1) % 3
    ii = np.concatenate((i, i), axis=1)
    ij = np.concatenate((i, j), axis=1)
    ik = np.concatenate((i, k), axis=1)
    jk = np.concatenate((j, k), axis=1)
    q[:, 0] = R[:, 2, 1] - R[:, 1, 2]
    q[:, 1] = R[:, 0, 2] - R[:, 2, 0]
    q[:, 2] = R[:, 1, 0] - R[:, 0, 1]
    q[:, 3] = 1 + trace
    q[mask, 0] = 1 - trace + 2 * R[ii]
    q[mask, 1] = R[np.flip(ij)] + R[ij]
    q[mask, 2] = R[np.flip(ik)] + R[ik]
    q[mask, 3] = R[np.flip(jk)] - R[jk]
    raise NotImplementedError("please verify this function implementation before usage")
    # Rxx, Ryx, Rzx, Rxy, Ryy, Rzy, Rxz, Ryz, Rzz = R.flat
    # K = np.array([
    #     [Rxx - Ryy - Rzz, 0, 0, 0],
    #     [Ryx + Rxy, Ryy - Rxx - Rzz, 0, 0],
    #     [Rzx + Rxz, Rzy + Ryz, Rzz - Rxx - Ryy, 0],
    #     [Ryz - Rzy, Rzx - Rxz, Rxy - Ryx, Rxx + Ryy + Rzz]]) / 3.0
    # eigvals, eigvecs = np.linalg.eigh(K)
    # qvec = eigvecs[[3, 0, 1, 2], np.argmax(eigvals)]
    # if qvec[0] < 0:
    #     qvec *= -1
    # return qvec
    return q


def mat2qvec(matrix):
    """
    Converts a batch of rotation matrices to a batch of quaternions
    :param matrix: tensor of size nx3x3
    :return: qvec: tensor of size nx4 where real part is first (a + bi + cj + dk) => (a, b, c, d)
    implementation is exact copy of pytorch3d implementation
    """
    if matrix.size(-1) != 3 or matrix.size(-2) != 3:
        raise ValueError(f"Invalid rotation matrix shape {matrix.shape}.")
    batch_dim = matrix.shape[:-2]
    m00, m01, m02, m10, m11, m12, m20, m21, m22 = torch.unbind(
        matrix.reshape(batch_dim + (9,)), dim=-1
    )
    q = torch.stack(
        [
            1.0 + m00 + m11 + m22,
            1.0 + m00 - m11 - m22,
            1.0 - m00 + m11 - m22,
            1.0 - m00 - m11 + m22,
        ],
        dim=-1,
    )
    q_abs = torch.zeros_like(q)
    positive_mask = q > 0
    q_abs[positive_mask] = torch.sqrt(q[positive_mask])
    # we produce the desired quaternion multiplied by each of r, i, j, k
    quat_by_rijk = torch.stack(
        [
            torch.stack([q_abs[..., 0] ** 2, m21 - m12, m02 - m20, m10 - m01], dim=-1),
            torch.stack([m21 - m12, q_abs[..., 1] ** 2, m10 + m01, m02 + m20], dim=-1),
            torch.stack([m02 - m20, m10 + m01, q_abs[..., 2] ** 2, m12 + m21], dim=-1),
            torch.stack([m10 - m01, m20 + m02, m21 + m12, q_abs[..., 3] ** 2], dim=-1),
        ],
        dim=-2,
    )
    # We floor here at 0.1 but the exact level is not important; if q_abs is small,
    # the candidate won't be picked.
    flr = torch.tensor(0.1).to(dtype=q_abs.dtype, device=q_abs.device)
    quat_candidates = quat_by_rijk / (2.0 * q_abs[..., None].max(flr))
    # if not for numerical problems, quat_candidates[i] should be same (up to a sign),
    # forall i; we pick the best-conditioned one (with the largest denominator)
    return quat_candidates[
        torch.nn.functional.one_hot(q_abs.argmax(dim=-1), num_classes=4) > 0.5, :
    ].reshape(batch_dim + (4,))



# shows some use cases of calibration and reconstruction
import gsoup
from pathlib import Path
import numpy as np

# instantiate a gray code object
gray = gsoup.GrayCode()
# set projector resolution
proj_wh = (800, 800)
# set decoding mode (rows first or columns first)
mode = "ij"  # "ij" is rows first, "xy" is columns first
# generate the gray code patterns, this is the ground truth
orig_patterns = gray.encode(proj_wh)
# save the patterns for projection and capturing
gsoup.save_images(orig_patterns, "tests/tests_resource/patterns/")
### start capture time ###
# here is the point you need to project these patterns onto the scene and capture with a camera
# the captured images should be saved in a folder with increasing numbers or increasing alphanumerical as names
### end capture time ###
# load simulated captured images
captured_patterns = gsoup.load_images(
    Path("tests/tests_resource/correspondence"),
)
# find camera resolution
cam_wh = (
    captured_patterns[0].shape[1],
    captured_patterns[0].shape[0],
)
# finally, decode the captured images
# the result is a dense map from camera pixels to projector pixels and a foreground mask
forward_map, fg = gray.decode(
    captured_patterns,
    (800, 800),
    output_dir="resource/forward",
    debug=True,
    mode=mode,
)
# the map is nice for many tasks, but we want to reconsutrct the scene in 3D, so we need to calibrate the procam pair.
# first, lets decide where the camera is. Usually you will set this as identity (camera is at the origin).
# but this scene was simulated in blender with a different known camera transform, so we will use that transform here.
blend_to_cv = np.array([[1.0, 0, 0, 0], [0, -1, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]])
cam_transform = np.array([[0, 0, 1, 1], [1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1.0]])
cam_transform = cam_transform @ blend_to_cv
### gt information: ###
# cam_int = np.array([[800, 0, 400.0], [0.0, 800, 400], [0, 0, 1]])
# proj_int = np.array([[800, 0, 400.0], [0.0, 800, 400], [0, 0, 1]])
# proj_transform = np.array([[-0.12403473, -0.23891242,  0.96308672,  0.8 ],
#                             [ 0.99227786, -0.02986405,  0.12038584,  0.1],
#                             [ 0.        ,  0.97058171,  0.24077168,  0.2],
#                             [ 0.        ,  0.        ,  0.        ,  1. ]])
# proj_transform = proj_transform @ blend_to_cv
# cam_dist = None
### end gt info. ###
# let's calibrate the procam pair. this requires capturing a chessboard pattern with the camera and projecting the patterns with the projector.
# see the folder for examples of the captured images.
result = gsoup.calibrate_procam(
    (800, 800),
    Path("tests/tests_resource/calibration"),
    chess_vert=15,
    chess_hori=15,
    chess_block_size=0.0185,
    output_dir="resource/calibration",
    projector_orientation="none",
    debug=True,
)
# the result is a dictionary with everything we need to reconstruct the scene
cam_int, cam_dist, proj_int, proj_dist, proj_transform = (
    result["cam_intrinsics"],
    result["cam_distortion"],
    result["proj_intrinsics"],
    result["proj_distortion"],
    result["proj_transform"],
)
# lets find the projector to world transform (the result above is the projector to camera transform)
proj_transform = cam_transform @ np.linalg.inv(proj_transform)  # p2w = c2w @ p2c

# we can visiualize the result using the viewer (commented out)
# calibration_static_view(cam_transform, proj_transform, (800, 800), (800, 800), cam_int, cam_dist, proj_int, forward_map, fg, mode)

# but more importantly, let us reconstruct a point cloud of scene
pc = gsoup.reconstruct_pointcloud(
    forward_map,
    fg,
    cam_transform,
    proj_transform,
    cam_int,
    cam_dist,
    proj_int,
    mode=mode,
)
# save the point cloud in a .ply format. this can be viewed in meshlab or other 3D viewers
gsoup.save_pointcloud(pc, "resource/points.ply")



import numpy as np
import cv2
from .gsoup_io import save_image, save_images, load_images, load_image
from .transforms import compose_rt
from .core import to_8b, to_hom, swap_columns
from .image import change_brightness
from .geometry_basic import ray_ray_intersection, point_line_distance
from pathlib import Path
from scipy.interpolate import LinearNDInterpolator
from scipy.spatial import ConvexHull


def blend_intensity_multi_projectors(forward_maps, fgs, proj_whs, mode="ij"):
    """
    given an image of the overlapping region of two or more projectors, and dense mappings between camera and all projectors pixels, computes an alpha mask per projector for seamless projection
    based on: "Multiprojector Displays using Camera-Based Registration".
    :param forward_maps: a list of forward maps each of (proj_hi x proj_wi x 2) uint32 corresponding to projectors i resolution
    :param fgs: a list of foreground masks each of (proj_hi x proj_wi) bool corresponding to projectors i resolution
    :proj_whs: a list of projector resolutions (proj_wi, proj_hi) as tuples
    :param mode: "xy" or "ij" depending on the order of the last channel of forward_map (see GrayCode.decode)
    :return: a list of alpha masks to use per projector i of size (proj_hi x proj_wi x 1) uint8
    """
    cam_wh = forward_maps[0].shape[:2]
    multi_proj_map = np.ones(
        (len(forward_maps), cam_wh[1], cam_wh[0], 1), dtype=np.float
    )
    for i in range(len(forward_maps)):
        points = np.where(fgs[i])
        convex_hull = ConvexHull(points)
        hull_points = [points[simplex] for simplex in convex_hull.simplices] + [
            points[convex_hull.simplices[0]]
        ]
        hull_edges = np.array(
            [[hull_points[i], hull_points[i + 1]] for i in range(len(hull_points) - 1)]
        )
        distances = np.empty(len(hull_edges), len(points))
        for j, edge in enumerate(hull_edges):
            distances[j] = point_line_distance(points, edge[0], edge[1])
        distances = distances.min(axis=0)
        multi_proj_map[i, points] = distances
    multi_proj_map = multi_proj_map / multi_proj_map.sum(axis=0, keepdims=True)
    multi_proj_map = to_8b(multi_proj_map)
    masks = []
    for fmap, fg, proj_wh in zip(forward_maps, fgs, proj_whs):
        bmap = compute_backward_map(proj_wh, fmap, fg, mode=mode)
        masks.append(multi_proj_map[bmap])
    return masks


def warp_image(backward_map, desired_image, cam_wh=None, mode="xy", output_path=None):
    """
    given a 2D dense map of corresponding pixels between projector and camera, computes a warped image such that if projected, the desired image appears when observed using camera
    :param backward_map: 2D dense mapping between pixels from projector 2 camera (proj_h x proj_w x 2) uint32
    :param desired_image: path to desired image from camera, or float np array channels last (cam_h x cam_w x 3) uint8
    :param cam_wh: device1 (width, height) as tuple, if not supplied assumes desired_image is in the correct dimensions in relation to backward_map
    :param mode: "xy" or "ij" depending on the last channel of backward_map
    :param output_path: path to save warped image to
    :return: warped image (proj_h x proj_w x 3) uint8
    """
    if backward_map.ndim != 3:
        raise ValueError("backward_map must be 3D")
    if backward_map.shape[-1] != 2:
        raise ValueError("backward_map must have shape (cam_h, cam_w, 2)")
    if type(desired_image) == np.ndarray:
        if desired_image.ndim != 3:
            raise ValueError("desired_image must be 3D")
    else:
        if cam_wh is not None:
            desired_image = load_image(desired_image, resize_wh=cam_wh)
        else:
            desired_image = load_image(desired_image)
    if mode == "xy":
        warpped = desired_image[(backward_map[..., 1], backward_map[..., 0])]
    elif mode == "ij":
        warpped = desired_image[(backward_map[..., 0], backward_map[..., 1])]
    else:
        raise ValueError("mode must be 'xy' or 'ij'")
    if output_path is not None:
        output_path = Path(output_path)
        output_path.parent.mkdir(parents=True, exist_ok=True)
        save_image(warpped, output_path)
    return warpped


def compute_backward_map(
    proj_wh,
    forward_map,
    foreground,
    mode="ij",
    interpolate=True,
    output_dir=None,
    debug=False,
):
    """
    computes the inverse map of forward_map by piece-wise interpolating a triangulated version of it (see GrayCode.decode to understand forward_map)
    :param proj_wh: projector (width, height) as a tuple
    :param forward_map: forward map as a numpy array of shape (height, width, 2) of type int32
    :param foreground: a boolean mask of shape (height, width) where True indicates a valid pixel in forward_map
    :param interpolate: if False, output will not be interpolated (will have holes...)
    :param mode: the last channel order of forward_map. either "xy" (width first) or "ij" (height first).
    :param output_dir: directory to save the inverse map to
    :param debug: if True, saves a visualization of the backward map to output_dir (R=X, G=Y, B=0) where X increases from left to right and Y increases from top to bottom
    :return: backward map as a numpy array of shape (projector_height, projector_width, 2) of type int32, last channel encodes (height, width).
    """
    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)
    data = np.argwhere(foreground)  # always ij
    points = forward_map[foreground]
    if interpolate:
        interp = LinearNDInterpolator(points, data, fill_value=0.0)
        X, Y = np.meshgrid(np.arange(proj_wh[0]), np.arange(proj_wh[1]))
        result = interp(X, Y).transpose(1, 0, 2)  # eww
    else:
        sparse_map = np.zeros((proj_wh[1], proj_wh[0], 2), dtype=np.uint32)
        if mode == "xy":
            sparse_map[(points[:, 1]), (points[:, 0])] = data
        elif mode == "ij":
            sparse_map[(points[:, 0]), (points[:, 1])] = data
        result = sparse_map
    if output_dir:
        np.save(Path(output_dir, "backward_map.npy"), result)
        if debug:
            result_normalized = result / np.array(
                [forward_map.shape[0], forward_map.shape[1]]
            )
            result_normalized[..., [0, 1]] = result_normalized[..., [1, 0]]
            result_normalized_8b = to_8b(result_normalized)
            result_normalized_8b_3c = np.concatenate(
                (result_normalized_8b, np.zeros_like(result_normalized_8b[..., :1])),
                axis=-1,
            )
            save_image(result_normalized_8b_3c, Path(output_dir, "backward_map.png"))
    return result.round().astype(np.uint32)


def naive_color_compensate(
    target_image,
    all_white_image,
    all_black_image,
    cam_width,
    cam_height,
    brightness_decrease=-127,
    projector_gamma=2.2,
    output_path=None,
    debug=False,
):
    """
    color compensate a projected image such that it appears closer to a target image from the perspective of a camera
    based on "Embedded entertainment with smart projectors".
    :param target_image the desired image path from the perspective of the camera
    :param all_white_image a path to picture taken by camera when projector had all pixels fully on (float32)
    :param all_black_image a path to picture taken by camera when projector had all pixels fully off (float32)
    :param cam_width camera image width
    :param cam_height camera image height
    :param brightness_decrease a hyper parameter controlling how much the total brightness is decreased. without this, the result is saturated because of dividing by small numbers
    :param projector_gamma under normal circumstances the projector does not actually output linear values, so we need to compensate for that
    :output_path if passed, result will be saved to this path
    :debug if true, will save debug info into output_path parent directory
    """
    if output_path is not None:
        output_path = Path(output_path)
        output_path.parent.mkdir(parents=True, exist_ok=True)
    target_image = load_image(
        target_image, to_float=True, resize_wh=(cam_width, cam_height)
    )[..., :3]
    target_image = change_brightness(target_image, brightness_decrease)
    if debug:
        save_image(target_image, Path(output_path.parent, "decrease_brightness.png"))
    all_white_image = load_image(
        all_white_image, to_float=True, resize_wh=(cam_width, cam_height)
    )
    all_black_image = load_image(
        all_black_image, to_float=True, resize_wh=(cam_width, cam_height)
    )
    compensated = (target_image - all_black_image) / all_white_image
    compensated = np.power(compensated, (1 / projector_gamma))
    compensated = np.nan_to_num(compensated, nan=0.0, posinf=0.0, neginf=0.0)
    compensated = np.clip(compensated, 0, 1)
    if output_path:
        save_image(compensated, output_path)
    return compensated


def calibrate_procam(
    proj_wh,
    capture_dir,
    chess_vert=10,
    chess_hori=7,
    bg_threshold=10,
    chess_block_size=10.0,
    projector_orientation="lower_half",
    verbose=True,
    output_dir=None,
    debug=False,
):
    """
    calibrates a projection-camera pair using local homographies
    based on "Simple, accurate, and robust projector-camera calibration".
    note1: the calibration poses some reasonable constraints on the projector-camera pair:
    1) projector is assumed to have no distortion, and a square pixel aspect ratio.
    2) camera is assumed to have a square pixel aspect ratio, and principle axis is assumed to be the center of the image.
    3) both elements are assumed to have no tangential distortion.
    note2: recommendations for calibration:
    1) use a chessboard with as many blocks as possible, but make sure the chessboard is fully visible in the camera image and in focus.
    2) capture sessions should span the whole image plane, and should be as diverse as possible in terms of poses (tilt the checkerboard !)
    3) when tilting the checkerboard, do not tilt it too much or the projector pixels will get smeared and the gray code decoding will produce large errors.
    4) place the checkeboard only in the working zone of the projector, i.e. the area where the projector is in focus.
    5) make sure the full dynamic range of the camera is used, i.e. the blackest black and whitest white should be visible in the captured images.
    6) attach the checkerboard to a flat surface, make sure the final pattern is as flat as possible. any bending will cause large errors.
    7) capture as many sessions as possible, if a session is not good (use debug=True to check), discard it.
    8) the RMS is only a rough estimate of the calibration quality, but if it is below 1, the calibration should be good enough (units are pixels).
    :param proj_wh projector resolution (width, height) as a tuple
    :param chess_vert when holding the chessboard in portrait mode, this is the number of internal "crossing points" of chessboard in vertical direction (i.e. where two white and two black squares meet)
    :param chess_hori when holding the chessboard in portrait mode, this is the number of internal "crossing points" of chessboard in horizontal direction (i.e. where two white and two black squares meet)
    :param capture_dir directory containing the captured images when gray code patterns were projected, assumes structure as follows:
        capture_dir
            - folder1
                - 0000.png
                - 0001.png
                - ...
            - folder2
                - 0000.png
                - ...
    :param bg_threshold threshold for detecting foreground
    :param chess_block_size size of blocks of chessboard in real world in any unit of measurement (will only effect the translation componenet between camera and projector)
    :param projector_orientation -  effects initial guess for projector principal point.
                                    can be "upper_half" or "lower_half", or "none".
                                    tabletop projectors usually have their principal point in the lower half of the image while ceiling mounted projectors have it in the upper half.
                                    note: opencv convention has the y direction increasing downwards.
    :param verbose if true, will print out the calibration results
    :param output_dir will save debug results to this directory
    :param debug if true, will save debug info into output_dir
    :return camera intrinsics (3x3),
            camera distortion parameters (5x1),
            projector intrinsics (3x3),
            projector distortion parameters (5x1),
            proj_transform (4x4), a camera to projector (c2p) transformation matrix (to get p2w, you should invert this matrix and multiply from the left with the camera to world matrix i.e. p2w = c2w * p2c)
    """
    chess_shape = (chess_vert, chess_hori)
    capture_dir = Path(capture_dir)
    if not capture_dir.exists():
        raise FileNotFoundError("capture_dir was not found")
    dirnames = sorted([x for x in capture_dir.glob("*") if x.is_dir()])
    if len(dirnames) == 0:
        raise FileNotFoundError("capture_dir contains no subfolders")
    used_dirnames = []
    gc_fname_lists = []
    for dname in dirnames:
        gc_fnames = sorted(dname.glob("*"))
        if len(gc_fnames) == 0:
            continue
        used_dirnames.append(str(dname))
        gc_fname_lists.append([str(x) for x in gc_fnames])
    dirnames = used_dirnames
    objps = np.zeros((chess_shape[0] * chess_shape[1], 3), np.float32)
    objps[:, :2] = chess_block_size * np.mgrid[
        0 : chess_shape[0], 0 : chess_shape[1]
    ].T.reshape(-1, 2)
    graycode = GrayCode()
    patterns = graycode.encode(proj_wh)
    cam_shape = load_image(gc_fname_lists[0][0], as_grayscale=True)[:, :, 0].shape[
        ::-1
    ]  # width, height
    patch_size_half = int(
        np.ceil(cam_shape[0] / 180)
    )  # some magic number for patch size
    cam_corners_list = (
        []
    )  # will contain the corners of the chessboard in camera coordinates
    cam_objps_list = (
        []
    )  # will contain the corners of the chessboard in chessboard local coordinates (unit of measurement deduced from chess_block_size)
    cam_corners_list2 = []
    proj_objps_list = []
    proj_corners_list = []
    for dname, gc_filenames in zip(dirnames, gc_fname_lists):
        if verbose:
            print("processing: {}".format(dname))
        if len(gc_filenames) != len(patterns):
            raise ValueError("invalid number of images in " + dname)
        imgs = load_images(gc_filenames, as_grayscale=True)
        forwardmap, fg = graycode.decode(
            imgs,
            proj_wh,
            mode="xy",
            bg_threshold=bg_threshold,
            output_dir=output_dir,
            debug=debug,
        )
        black_img = imgs[-1]
        white_img = imgs[-2]
        imgs = imgs[:-2]
        res, cam_corners = cv2.findChessboardCorners(white_img, chess_shape)
        if not res:
            raise RuntimeError(
                "chessboard was not found in {}".format(gc_filenames[-2])
            )
        cam_objps_list.append(objps)
        cam_corners_list.append(cam_corners)
        proj_objps = []
        proj_corners = []
        cam_corners2 = []
        for corner, objp in zip(cam_corners, objps):
            c_x = int(round(corner[0][0]))
            c_y = int(round(corner[0][1]))
            src_points = []
            dst_points = []
            # todo: vectorize these loops
            for dx in range(-patch_size_half, patch_size_half + 1):
                for dy in range(-patch_size_half, patch_size_half + 1):
                    x = c_x + dx
                    y = c_y + dy
                    if fg[y, x]:
                        proj_pix = forwardmap[y, x]
                        src_points.append((x, y))
                        dst_points.append(np.array(proj_pix))
            if len(src_points) < patch_size_half**2:
                if verbose:
                    print(
                        "corner {}, {} was skiped because too few decoded pixels found (check your images and thresholds)".format(
                            c_x, c_y
                        )
                    )
                continue
            h_mat, inliers = cv2.findHomography(
                np.array(src_points), np.array(dst_points)
            )
            point = h_mat @ np.array([corner[0][0], corner[0][1], 1]).transpose()
            point_pix = point[0:2] / point[2]
            proj_objps.append(objp)
            proj_corners.append([point_pix])
            cam_corners2.append(corner)
        if len(proj_corners) < 3:
            raise RuntimeError(
                "too few corners were found in {} (less than 3)".format(dname)
            )
        proj_objps_list.append(np.float32(proj_objps))
        proj_corners_list.append(np.float32(proj_corners))
        cam_corners_list2.append(np.float32(cam_corners2))
    if verbose:
        print(
            "total correspondence points: {}".format(
                sum([len(x) for x in proj_corners_list])
            )
        )
    # Initial solution of camera's intrinsic parameters
    # camera_intrinsics_init = np.array([[np.mean(cam_shape), 0, cam_shape[0]/2], [0, np.mean(cam_shape), cam_shape[1]/2], [0, 0, 1]])
    ret, cam_int, cam_dist, cam_rvecs, cam_tvecs = cv2.calibrateCamera(
        cam_objps_list,
        cam_corners_list,
        cam_shape,
        None,
        None,
        None,
        None,  # camera_intrinsics_init
        cv2.CALIB_FIX_ASPECT_RATIO + cv2.CALIB_FIX_PRINCIPAL_POINT,
    )  # + cv2.CALIB_ZERO_TANGENT_DIST  # cv2.CALIB_USE_INTRINSIC_GUESS
    if verbose:
        print("Camera calib. intrinsic parameters: {}".format(cam_int))
        print("Camera calib. distortion parameters: {}".format(cam_dist))
        print("Camera calib. reprojection error: {}".format(ret))

    # Initial solution of projector's parameters
    if projector_orientation == "none":
        cy_correction = 0
    elif projector_orientation == "lower_half":
        cy_correction = proj_wh[1] / 4
    elif projector_orientation == "upper_half":
        cy_correction = -proj_wh[1] / 4
    else:
        raise ValueError("invalid projector_orientation")
    projector_intrinsics_init = np.array(
        [
            [np.mean(proj_wh), 0, proj_wh[0] / 2],
            [0, np.mean(proj_wh), cy_correction + proj_wh[1] / 2],
            [0, 0, 1],
        ]
    )
    projector_ditortion_init = np.zeros((5, 1))
    ret, proj_int, proj_dist, proj_rvecs, proj_tvecs = cv2.calibrateCamera(
        proj_objps_list,
        proj_corners_list,
        proj_wh,
        projector_intrinsics_init,
        projector_ditortion_init,
        None,
        None,
        cv2.CALIB_USE_INTRINSIC_GUESS
        + cv2.CALIB_FIX_ASPECT_RATIO
        + cv2.CALIB_ZERO_TANGENT_DIST
        + cv2.CALIB_FIX_K1
        + cv2.CALIB_FIX_K2
        + cv2.CALIB_FIX_K3,
    )

    if verbose:
        print("Projector calib. intrinsic parameters: {}".format(proj_int))
        print("Projector calib. distortion parameters: {}".format(proj_dist))
        print("Projector calib. reprojection error: {}".format(ret))

    # Stereo calibration for final solution
    (
        ret,
        cam_int,
        cam_dist,
        proj_int,
        proj_dist,
        cam_proj_rmat,
        cam_proj_tvec,
        E,
        F,
    ) = cv2.stereoCalibrate(
        proj_objps_list,
        cam_corners_list2,
        proj_corners_list,
        cam_int,
        cam_dist,
        proj_int,
        proj_dist,
        None,
    )

    proj_transform = compose_rt(
        cam_proj_rmat[None, ...], cam_proj_tvec[None, :, 0], square=True
    )[0]
    if verbose:
        print("Stereo reprojection error: {}".format(ret))
        print("Stereo camera intrinsic parameters: {}".format(cam_int))
        print("Stereo camera distortion parameters: {}".format(cam_dist))
        print("Stereo projector intrinsic parameters: {}".format(proj_int))
        print("Stereo projector distortion parameters: {}".format(proj_dist))
        print("Stereo camera2projector transform): {}".format(proj_transform))
    if debug:
        # computes a histogram of camera reprojection errors, and not just average error
        cam_corners = np.array(cam_corners_list).squeeze()
        proj_corners = np.array(proj_corners_list).squeeze()
        obj_corners = np.array(cam_objps_list).squeeze()
        all_projected_cam_corners = []
        all_projected_proj_corners = []
        for i in range(len(cam_corners)):
            projected_cam_points, _ = cv2.projectPoints(
                obj_corners[i], cam_rvecs[i], cam_tvecs[i], cam_int, cam_dist
            )
            all_projected_cam_corners.append(projected_cam_points)
            projected_proj_points, _ = cv2.projectPoints(
                obj_corners[i], proj_rvecs[i], proj_tvecs[i], proj_int, proj_dist
            )
            all_projected_proj_corners.append(projected_proj_points)
        all_projected_cam_corners = np.array(all_projected_cam_corners).squeeze()
        cam_norms = np.linalg.norm(cam_corners - all_projected_cam_corners, axis=-1)
        cam_per_session_error = cam_norms.mean(axis=-1)
        worst_to_best_cam_session_ids = np.argsort(cam_per_session_error)[::-1]
        worst_to_best_cam_errors = cam_per_session_error[worst_to_best_cam_session_ids]
        print(
            "worst to best sessions ids for camera reprojection error: {}".format(
                worst_to_best_cam_session_ids
            )
        )
        print("and their associated errors: {}".format(worst_to_best_cam_errors))
        cam_hist = np.histogram(cam_norms)
        print(
            "camera reprojection error histogram: {} (should be similar to gaussian around 0)".format(
                cam_hist
            )
        )
        all_projected_proj_corners = np.array(all_projected_proj_corners).squeeze()
        proj_norms = np.linalg.norm(proj_corners - all_projected_proj_corners, axis=-1)
        per_session_projector_error = proj_norms.mean(axis=-1)
        worst_to_best_proj_session_ids = np.argsort(per_session_projector_error)[::-1]
        worst_to_best_proj_errors = per_session_projector_error[
            worst_to_best_proj_session_ids
        ]
        print(
            "worst to best sessions ids for projector reprojection error: {}".format(
                worst_to_best_proj_session_ids
            )
        )
        print("and their associated errors: {}".format(worst_to_best_proj_errors))
        proj_hist = np.histogram(proj_norms)
        print(
            "projector reprojection error histogram: {} (should be similar to gaussian around 0)".format(
                proj_hist
            )
        )
    ret = {
        "cam_intrinsics": cam_int,
        "cam_distortion": cam_dist,
        "proj_intrinsics": proj_int,
        "proj_distortion": proj_dist,
        "proj_transform": proj_transform,
    }
    return ret


def reconstruct_pointcloud(
    forward_map,
    fg,
    cam_transform,
    proj_transform,
    cam_int,
    cam_dist,
    proj_int,
    mode="ij",
    color_image=None,
    debug=False,
):
    """
    given a dense pixel correspondence map between a camera and a projector, and calibration results, reconstructs a 3D point cloud of the scene.
    :param forward_map: a dense pixel correspondence map between a camera and a projector (see GrayCode.decode)
    :param fg: a foreground mask of the scene (see GrayCode.decode)
    :param cam_transform: a 4x4 transformation matrix of the camera (cam to world)
    :param proj_transform: a 4x4 transformation matrix of the projector (proj to world)
    :param cam_int: camera's intrinsic parameters
    :param cam_dist: camera's distortion parameters
    :param proj_int: projector's intrinsic parameters
    :param mode: "xy" or "ij" which is the ordering of the last channel of the forward map (see GrayCode.decode)
    :param color_image: RGB image with the same spatial size as forwardmap. if supplied will return a colored point cloud (Nx6).
    :param debug: if True, will return debug information
    :return: a 3D point cloud of the scene (Nx3)
    """
    cam_pixels = swap_columns(np.argwhere(fg), 0, 1)
    # todo: account for distortion if supplied
    # undistorted_cam_points =  cv2.undistortPoints(cam_points, cam_int, cam_dist, P=proj_transform[0]).squeeze()  # equivalent to not setting P and doing K @ points outside
    cam_origins = cam_transform[None, :3, -1]
    cam_directions = (
        cam_transform[:3, :3] @ (np.linalg.inv(cam_int) @ to_hom(cam_pixels).T)
    ).T
    cam_directions = cam_directions / np.linalg.norm(
        cam_directions, axis=-1, keepdims=True
    )
    if mode == "xy":
        projector_pixels = forward_map[fg]
    elif mode == "ij":
        projector_pixels = swap_columns(forward_map[fg], 0, 1)
    else:
        raise ValueError("mode must be either 'xy' or 'ij'")
    projector_origins = proj_transform[None, :3, -1]
    projector_directions = (
        proj_transform[:3, :3] @ (np.linalg.inv(proj_int) @ to_hom(projector_pixels).T)
    ).T
    projector_directions = projector_directions / np.linalg.norm(
        projector_directions, axis=-1, keepdims=True
    )
    # note: using ray-ray is slightly naive. A better way is to minimize a reprojection error, and then ray-ray
    points, weight_factor = ray_ray_intersection(
        cam_origins, cam_directions, projector_origins, projector_directions
    )
    if color_image is not None:
        colors = (
            color_image[fg] if mode == "xy" else color_image[swap_columns(fg, 0, 1)]
        )
        points = np.concatenate([points, colors], axis=-1)
    if debug:
        return (
            points,
            weight_factor,
            cam_origins,
            cam_directions,
            projector_origins,
            projector_directions,
            cam_pixels,
            projector_pixels,
        )
    else:
        return points


class GrayCode:
    """
    a class that handles encoding and decoding graycode patterns
    """

    def encode1d(self, length):
        total_images = len(bin(length - 1)[2:])

        def xn_to_gray(n, x):
            # infer a coordinate gray code from its position x and index n (the index of the image out of total_images)
            # gray code is obtained by xoring the bits of x with itself shifted, and selecting the n-th bit
            return (x ^ (x >> 1)) & (1 << (total_images - 1 - n)) != 0

        imgs_code = 255 * np.fromfunction(
            xn_to_gray, (total_images, length), dtype=int
        ).astype(np.uint8)
        return imgs_code

    def encode(self, proj_wh, flipped_patterns=True):
        """
        encode projector's width and height into gray code patterns
        :param proj_wh: projector's (width, height) in pixels as a tuple
        :param flipped_patterns: if True, flipped patterns are also generated for better binarization
        :return: a 3D numpy array of shape (total_images, height, width) where total_images is the number of gray code patterns
        """
        width, height = proj_wh
        codes_width_1d = self.encode1d(width)[:, None, :]
        codes_width_2d = codes_width_1d.repeat(height, axis=1)
        codes_height_1d = self.encode1d(height)[:, :, None]
        codes_height_2d = codes_height_1d.repeat(width, axis=2)

        img_white = np.full((height, width), 255, dtype=np.uint8)[None, ...]
        img_black = np.full((height, width), 0, dtype=np.uint8)[None, ...]
        all_images = np.concatenate((codes_width_2d, codes_height_2d), axis=0)
        if flipped_patterns:
            all_images = np.concatenate((all_images, 255 - codes_width_2d), axis=0)
            all_images = np.concatenate((all_images, 255 - codes_height_2d), axis=0)
        all_images = np.concatenate((all_images, img_white, img_black), axis=0)
        return all_images[..., None]

    def binarize(self, captures, flipped_patterns=True, bg_threshold=10):
        """
        binarize a batch of images
        :param captures: a 4D numpy array of shape (n, height, width, 1) of captured images
        :param flipped_patterns: if true, patterns also contain their flipped version for better binarization
        :param bg_threshold: if the difference between the pixel in the white&black images is greater than this, pixel is foreground
        :return: a 4D numpy binary array for decoding (total_images, height, width, 1) where total_images is the number of gray code patterns
        and a binary foreground mask (height, width, 1)
        """
        if not -255 <= bg_threshold <= 255:
            raise ValueError("bg_threshold must be between -255 and 255")
        patterns, bw = captures[:-2], captures[-2:]
        foreground = bw[0].astype(np.int32) - bw[1].astype(np.int32) > bg_threshold
        if flipped_patterns:
            orig, flipped = (
                patterns[: len(patterns) // 2],
                patterns[len(patterns) // 2 :],
            )
            # valid = (orig.astype(np.int32) - flipped.astype(np.int32)) > bin_threshold
            binary = orig > flipped
            # foreground = foreground & np.all(valid, axis=0)  # only pixels that are valid in all images are foreground
        else:  # slightly more naive thresholding
            binary = patterns >= 0.5 * (bw[1] + bw[0])
        return binary, foreground

    def decode1d(self, gc_imgs):
        # gray code to binary
        n, h, w = gc_imgs.shape
        binary_imgs = gc_imgs.copy()
        for i in range(1, n):  # xor with previous image except MSB
            binary_imgs[i, :, :] = np.bitwise_xor(
                binary_imgs[i, :, :], binary_imgs[i - 1, :, :]
            )
        # decode binary
        cofficient = np.fromfunction(
            lambda i, y, x: 2 ** (n - 1 - i), (n, h, w), dtype=int
        )
        img_index = np.sum(binary_imgs * cofficient, axis=0)
        return img_index

    def decode(
        self,
        captures,
        proj_wh,
        flipped_patterns=True,
        bg_threshold=10,
        mode="ij",
        output_dir=None,
        debug=False,
    ):
        """
        decodes a batch of images encoded with gray code
        :param captures: a 4D numpy array of shape (n, height, width, c) of captured images (camera resolution is inferred from this)
        :param flipped_patterns: if true, patterns also contain their flipped version for better binarization
        :param bg_threshold: a threshold used for background detection using the all-white and all-black captures
        :param mode: "xy" or "ij" decides the order of last dimension coordinates in the output (ij -> height first, xy -> width first)
        :param output_dir: if not None, saves the decoded images to this directory
        :param debug: if True, visualizes the map in an image using the red and green channels where R=X, G=Y, B=0, X increases from left to right, Y increases from top to bottom
        :return: a 2D numpy array of shape (height, width, 2) uint32 mapping from camera pixels to projector pixels
        and a foreground mask (height, width) of booleans
        """
        if captures.ndim != 4:
            raise ValueError("captures must be a 4D numpy array")
        if captures.dtype != np.uint8:
            raise ValueError("captures must be uint8")
        if output_dir is not None:
            output_dir = Path(output_dir)
            output_dir.mkdir(parents=True, exist_ok=True)
        b, _, _, c = captures.shape
        b = b - 2  # dont count white and black images
        if flipped_patterns:
            b = b // 2  # dont count flipped patterns
        encoded = self.encode(
            proj_wh, flipped_patterns
        )  # sanity: encode with same arguments to verify enough captures are present
        if len(encoded) != len(captures):
            raise ValueError("captures must have length of {}".format(len(encoded)))
        if c != 1:  # naively convert to grayscale
            captures = captures.mean(axis=-1, keepdims=True).round().astype(np.uint8)
        imgs_binary, fg = self.binarize(captures, flipped_patterns, bg_threshold)
        imgs_binary = imgs_binary[:, :, :, 0]
        fg = fg[:, :, 0]
        x = self.decode1d(imgs_binary[: b // 2])
        y = self.decode1d(imgs_binary[b // 2 :])
        if mode == "ij":
            forward_map = np.concatenate((y[..., None], x[..., None]), axis=-1)
        elif mode == "xy":
            forward_map = np.concatenate((x[..., None], y[..., None]), axis=-1)
        else:
            raise ValueError("mode must be 'ij' or 'xy'")
        if output_dir is not None:
            np.save(Path(output_dir, "forward_map.npy"), forward_map)
            np.save(Path(output_dir, "fg.npy"), fg)
            if debug:
                save_images(imgs_binary[..., None], Path(output_dir, "imgs_binary"))
                save_image(fg, Path(output_dir, "foreground.png"))
                composed = forward_map * fg[..., None]
                if mode == "ij":
                    composed_normalized = composed / np.array([proj_wh[1], proj_wh[0]])
                    composed_normalized[..., [0, 1]] = composed_normalized[..., [1, 0]]
                elif mode == "xy":
                    composed_normalized = composed / np.array([proj_wh[0], proj_wh[1]])
                composed_normalized_8b = to_8b(composed_normalized)
                composed_normalized_8b_3c = np.concatenate(
                    (
                        composed_normalized_8b,
                        np.zeros_like(composed_normalized_8b[..., :1]),
                    ),
                    axis=-1,
                )
                save_image(
                    composed_normalized_8b_3c, Path(output_dir, "forward_map.png")
                )
        return forward_map, fg
